/*
-----------------------------------------------------------------------------

				IPDB additional information format library

											95/ 5/12	programmed by K.A

-----------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include "btype.h"
#include "handle.h"
#include "ipdbsys.h"
#include "aif.h"

enum{
	XYByte = 1, XYWord = 2, XYDWord = 3,
	XSign = 0, XUnsign = 0x4, XMinus = 0xc,
	YSign = 0, YUnsign = 0x10, YMinus = 0x30
};

enum{
	HT_AIF = 200
};

/* t֌W */
typedef struct tagITEM_LIST{
	char *item_name;
	int type, array_count;
	struct tagITEM_LIST *next;
} ITEM_LIST;

typedef struct{
	HANDLE_LIST handle_info;
	int aisize;
	ITEM_LIST *item_list;
} AI_FORMAT;

static int _aif_err_no = 0;

static AI_FORMAT *getaif(HAIFMT haif){
	return (AI_FORMAT *)handletopt(haif, HT_AIF);
}

static void stringcopy(char *dest, const char *src, int n){
	int i;
	char ch;
	n--;
	for(i = 0; i < n && (ch = src[i]) != 0; i++)
		dest[i] = ch;
	dest[i] = 0;
}

/* ̕ttH[}bg쐬 */
HAIFMT CreateAddInfoFormat(void){
	AI_FORMAT *aif;
	_aif_err_no = 0;
	if((aif = makehandleinfo(HT_AIF, sizeof(*aif))) == NULL)
		return (int)NULL;
	aif->item_list = NULL;
	aif->aisize = 0;
	return aif->handle_info.handle;
}

static void freeitemlist(ITEM_LIST *il){
	ITEM_LIST *tmp;
	while(il != NULL){
		tmp = il->next;
		free(il->item_name);
		free(il);
		il = tmp;
	}
}

/* ttH[}bgp */
void ReleaseAddInfoFormat(HAIFMT haif){
	AI_FORMAT *aif;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return;
	}
	freeitemlist(aif->item_list);
	removehandleinfo(haif, HT_AIF);
}

static int gettypesize(int type){
	int size = 0;
	switch(type){
	case IT_Char:
	case IT_Byte:
		size = sizeof(CHAR);
		break;
	case IT_Word:
		size = sizeof(WORD);
		break;
	case IT_DWord:
		size = sizeof(LONG);
		break;
	}
	return size;
}

static void setaiitem(void *dst, int type, DWORD src){
	switch(type){
	case IT_Char:
	case IT_Byte:
		*(CHAR *)dst = (CHAR)(src & 0xff);
		break;
	case IT_Word:
		*(SHORT *)dst = (SHORT)(src & 0xffff);
		break;
	case IT_DWord:
		*(LONG *)dst = src;
		break;
	}
}

static unsigned long getaiitem(const void *src, int type){
	LONG dst;
	switch(type){
	case IT_Char:
		dst = *(BYTE *)src;
		break;
	case IT_Byte:
		dst = *(CHAR *)src;
		break;
	case IT_Word:
		dst = *(SHORT *)src;
		break;
	case IT_DWord:
		dst = *(LONG *)src;
		break;
	}
	return dst;
}

void TransferAddInfo(const void *src, int src_offset, int src_type,
		int src_count, void *dst, int dst_offset, int dst_type, int dst_count){
	int i, ac = min(src_count, dst_count);
	int src_size = gettypesize(src_type), dst_size = gettypesize(dst_type);
	_aif_err_no = 0;
	if(!src_size || !dst_size){
		_aif_err_no = IPDB_IllegalValue;
		return;
	}
	(BYTE *)src += src_offset;
	(BYTE *)dst += dst_offset;
	for(i = 0; i < ac; i++){
		setaiitem(dst, dst_type, getaiitem(src, src_type));
		(BYTE *)src += src_size;
		(BYTE *)dst += dst_size;
	}
	if(src_count < dst_count)
		memset(dst, 0, dst_size);
}

static int getaiitemcount(AI_FORMAT *aif){
	int num = 0;
	ITEM_LIST *il;
	for(il = aif->item_list; il != NULL; il = il->next)
		num++;
	return num;
}

static int getitemsize(int type, int array_count){
	return gettypesize(type) * array_count;
}

static ITEM_LIST **ntoitempt(AI_FORMAT *aif, int item_No){
	int i;
	ITEM_LIST *tmp, **prev = &(aif->item_list);
	_aif_err_no = 0;
	if(item_No == NULL || item_No < 0){
		_aif_err_no = IPDB_IllegalValue;
		return NULL;
	}
	for(i = 1; (tmp = *prev) != NULL && i < item_No; i++)
		prev = &(tmp->next);
	if(tmp == NULL){
		_aif_err_no = IPDB_IllegalValue;
		return NULL;
	}
	return prev;
}

static ITEM_LIST *ntoitem(AI_FORMAT *aif, int item_No){
	ITEM_LIST **il;
	if((il = ntoitempt(aif, item_No)) == NULL)
		return NULL;
	return *il;
}

/* tϊ */
int ConvertAddInfo(HAIFMT src_haif, void *src, HAIFMT dst_haif, void *dst){
	void *dst_cur;
	AI_FORMAT *src_aif, *dst_aif;
	ITEM_LIST *src_il, *dst_il;

	_aif_err_no = 0;
	if((src_aif = getaif(src_haif)) == NULL ||
			(dst_aif = getaif(dst_haif)) == NULL)
		return (_aif_err_no = IPDB_IllegalHandle);
	for(src_il = src_aif->item_list; src_il != NULL; src_il = src_il->next){
		dst_il = dst_aif->item_list;
		dst_cur = dst;
		while(dst_il != NULL){
			if(!strcmp(dst_il->item_name, src_il->item_name))
				break;
			(BYTE *)dst_cur += getitemsize(dst_il->type, dst_il->array_count);
			dst_il = dst_il->next;
		}
		if(dst_il != NULL)
			TransferAddInfo(src, 0, src_il->type, src_il->array_count,
					dst_cur, 0, dst_il->type, dst_il->array_count);
		(BYTE *)src += getitemsize(src_il->type, src_il->array_count);
	}
	return 0;
}

static int itemton(ITEM_LIST *il, const char *item_name){
	int i;
	_aif_err_no = 0;
	for(i = 1; il != NULL && strcmp(il->item_name, item_name); i++)
		il = il->next;
	if(il == NULL){
		_aif_err_no = IPDB_IllegalName;
		return 0;
	}
	return i;
}

/* t񍀖ږ𓾂 */
int GetAddInfoItemName(HAIFMT haif,
		int item_No, char name_buf[], int buf_size){
	AI_FORMAT *aif;
	ITEM_LIST *il;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL)
		return (_aif_err_no = IPDB_IllegalHandle);
	if((il = ntoitem(aif, item_No)) == NULL)
		return _aif_err_no;
	stringcopy(name_buf, il->item_name, buf_size);
	return 0;
}

/* t̃TCY𓾂 */
int GetAddInfoSize(HAIFMT haif){
	AI_FORMAT *aif;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return aif->aisize;
}

/* t񍀖ڐ𓾂 */
int GetAddInfoItemCount(HAIFMT haif){
	AI_FORMAT *aif;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return getaiitemcount(aif);
}

/* t񍀖ڂ̔ԍ𓾂 */
int GetAddInfoItemNo(HAIFMT haif, const char item_name[]){
	AI_FORMAT *aif;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return itemton(aif->item_list, item_name);
}

static int addaiitem(AI_FORMAT *aif,
		const char item_name[], int type, int array_count){
	ITEM_LIST **il, *tmp;
	_aif_err_no = 0;
	if(type < IT_Byte || type > IT_Char || array_count <= 0)
		return (_aif_err_no = IPDB_IllegalValue);
	if(!item_name[0])
		return (_aif_err_no = IPDB_IllegalName);
	il = &(aif->item_list);
	while((tmp = *il) != NULL){
		if(!strncmp(tmp->item_name, item_name, MaxAIItemNameLen - 1))
			return (_aif_err_no = IPDB_AlreadyExist);
		il = &(tmp->next);
	}
	if((tmp = malloc(sizeof(ITEM_LIST))) == NULL)
		return (_aif_err_no = IPDB_MemoryAllocError);
	if((tmp->item_name =
			malloc(min(MaxAIItemNameLen, strlen(item_name) + 1))) == NULL){
		free(tmp);
		return (_aif_err_no = IPDB_MemoryAllocError);
	}
	tmp->type = type;
	tmp->array_count = array_count;
	tmp->next = NULL;
	stringcopy(tmp->item_name, item_name, MaxAIItemNameLen);
	*il = tmp;
	aif->aisize += getitemsize(type, array_count);
	return 0;
}

/* t񍀖ڂǉ */
int AddAddInfoItem(HAIFMT haif,
		const char *item_name, int type, int array_count){
	AI_FORMAT *aif;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL)
		return (_aif_err_no = IPDB_IllegalHandle);
	return addaiitem(aif, item_name, type, array_count);
}

/* t񍀖ڂ폜 */
int DeleteAddInfoItem(HAIFMT haif, int item_No){
	int type, array_count;
	AI_FORMAT *aif;
	ITEM_LIST **il, *tmp;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL)
		return (_aif_err_no = IPDB_IllegalHandle);
	if((il = ntoitempt(aif, item_No)) == NULL)
		return _aif_err_no;
	free((*il)->item_name);
	tmp = (*il)->next;
	type = (*il)->type;
	array_count = (*il)->array_count;
	free(*il);
	*il = tmp;
	aif->aisize -= getitemsize(type, array_count);
	return 0;
}

/* t񍀖ڃf[^̐擪ItZbg𓾂 */
int GetAddInfoItemOffset(HAIFMT haif, int item_No){
	AI_FORMAT *aif;
	ITEM_LIST *il;
	int offset = 0, n;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return 0;
	}
	if(item_No < 1){
		_aif_err_no = IPDB_IllegalValue;
		return 0;
	}
	il = aif->item_list;
	for(n = 1; n < item_No && il != NULL; n++){
		offset += getitemsize(il->type, il->array_count);
		il = il->next;
	}
	if(il == NULL){
		_aif_err_no = IPDB_IllegalValue;
		return 0;
	}
	return offset;
}

/* t񍀖ڃf[^̃TCY𓾂 */
int GetAddInfoItemSize(HAIFMT haif, int item_No){
	AI_FORMAT *aif;
	ITEM_LIST *il;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return 0;
	}
	if((il = ntoitem(aif, item_No)) == NULL)
		return 0;
	return getitemsize(il->type, il->array_count);
}

/* t񍀖ڃf[^̌^𓾂 */
int GetAddInfoItemType(HAIFMT haif, int item_No){
	AI_FORMAT *aif;
	ITEM_LIST *il;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return NULL;
	}
	if((il = ntoitem(aif, item_No)) == NULL)
		return NULL;
	return il->type;
}

/* t񍀖ڃf[^̔z񐔂𓾂 */
int GetAddInfoItemArrayCount(HAIFMT haif, int item_No){
	AI_FORMAT *aif;
	ITEM_LIST *il;
	_aif_err_no = 0;
	if((aif = getaif(haif)) == NULL){
		_aif_err_no = IPDB_IllegalHandle;
		return 0;
	}
	if((il = ntoitem(aif, item_No)) == NULL)
		return 0;
	return il->array_count;
}

/* ڂG[bZ[W𓾂 */
int GetAIFError(void){
	return _aif_err_no;
}
