/*
=============================================================================

				IPDB Library Version 2.12	۹۽߰ع EX ްޮ

											Programmed by Katsuhiko Akiyama

=============================================================================
C
ver 2.11 : 32 rbgΉ
ver 2.12 : [hI[t@CɑΉ

=============================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include "btype.h"
#include "handle.h"
#include "ipdbsys.h"

#define MaxPageCount			0xffff

enum{
	HT_IPDB = 1,
	ReadAcc = 1, WriteAcc, UpdateAcc
};

typedef struct tagINDEX_INFO{
	char name[IPDB_MaxIndexNameLen];
	WORD first_page, page_count;
	struct tagINDEX_INFO *next;
} INDEX_INFO;

typedef struct{
	FILE *fp;
	int acc_mode, prev_acc;
	int index_No;
	long offset, size;
	unsigned page_No;
	INDEX_INFO *index;
} PAGE_INFO;

typedef struct tagIPDB_INFO{
	HANDLE_LIST handle_info;
	char data_file[MAXPATH];
	unsigned page_count;
	INDEX_INFO *index;
	PAGE_INFO page;
	long flag_size;
	struct tagIPDB_INFO *chain;
} IPDB_INFO;

typedef struct{
	LONG size, index;
} PAGE_INDEX;

static const char _headerstring[] = "IPDB3.0a";
static const char _headerstring20[] = "IPDB2.0p";
static const char _aif_name[] = "additional information format";
static const char _set_ai_name[] = "set additional information";
static const char _opdb_name[] = "online pattern data";

#define HeadDataSize		(sizeof(_headerstring) - 1 + sizeof(DWORD) * 2)

static IPDB_INFO *_ipdb_info_chain = NULL;
static int _ipdb_err_no = 0;

#define CRC_SEED			0x11021L

static unsigned crccalc(void *data, long size, unsigned crc){
	int i;
	unsigned char next;
	unsigned long cur = 0, digit;
	for(digit = 0x80000000L; digit > CRC_SEED; digit >> 1);
	while(size > 0){
		next = *(((BYTE *)data)++);
		for(i = 0; i < 8; i++){
			if(cur & digit)
				cur ^= CRC_SEED;
			cur = (cur << 1) | ((next & 0x80) >> 7);
			next <<= 1;
		}
		size--;
	}
	for(i = 0; i < 16; i++){
		if(cur & digit)
			cur ^= CRC_SEED;
		cur = (cur << 1) | ((crc & 0x8000) >> 15);
		crc <<= 1;
	}
	return (unsigned)cur;
}

static IPDB_INFO *getipdbinfo(HIPDB hipdb){
	IPDB_INFO *ip;
	if((ip = handletopt(hipdb, HT_IPDB)) == NULL)
		return NULL;
	return ip;
}

static void *allocsorttbl(void *data, int data_size, long num){
	long i;
	void **pt, **tmp;
	if((pt = malloc(sizeof(void *) * num)) == NULL)
		return NULL;
	for(i = 0, tmp = pt; i < num; i++, (BYTE *)data += data_size, tmp++)
		*tmp = data;
	return pt;
}

static void msortsub(void **in,
		void **out, unsigned num, unsigned radix, int key_pos){
	unsigned i, j, n, nn, n1, n2, max1, max2;
	long d1, d2;
	void **cp1, **cp2, **wp;
	n = radix * 2;
	for(j = 0; j + radix < num; j += n){
		cp1 = &(in[j]);
		cp2 = &(in[j + radix]);
		n1 = n2 = 0;
		max1 = radix;
		max2 = (j + n >= num) ? num - (j + radix): radix;
		d1 = *(long *)((BYTE *)(*cp1) + key_pos);
		d2 = *(long *)((BYTE *)(*cp2) + key_pos);
		nn = (j + n >= num) ? num - j: n;
		for(i = 0, wp = &(out[j]); i < nn; i++, wp++){
			if(n1 >= max1 || (n2 < max2 && d1 > d2)){
				*wp = *(cp2++);
				if(++n2 < max2)
					d2 = *(long *)((BYTE *)(*cp2) + key_pos);
			}else{
				*wp = *(cp1++);
				if(++n1 < max1)
					d1 = *(long *)((BYTE *)(*cp1) + key_pos);
			}
		}
	}
	if(j < num){
		cp1 = &in[j];
		for(; j < num; j++)
			*wp++ = *cp1++;
	}
}

static void *msort(void *data, int data_size, unsigned num, int key_pos){
	unsigned i;
	void *tmp, **work, **stbl;

	_ipdb_err_no = 0;
	if((stbl = allocsorttbl(data, data_size, num)) == NULL){
		_ipdb_err_no = IPDB_MemoryAllocError;
		return NULL;
	}
	if((work = malloc(sizeof(*work) * num)) == NULL){
		free(stbl);
		_ipdb_err_no = IPDB_MemoryAllocError;
		return NULL;
	}
	for(i = 1; i < num; i += i){
		msortsub(stbl, work, num, i, key_pos);
		tmp = stbl;
		stbl = work;
		work = tmp;
	}
	free(work);
	return stbl;
}

static int getfilenum(const char *fn){
	int cnt;
	struct ffblk fb;
	if(findfirst(fn, &fb, FA_ARCH | FA_RDONLY) < 0)
		return 0;
	for(cnt = 0;;){
		cnt++;
		if(findnext(&fb) < 0)
			break;
	}
	return cnt;
}

static long getfilesize(const char fn[]){
	struct ffblk fb;
	if(findfirst(fn, &fb, FA_ARCH | FA_RDONLY))
		return -1;
	return fb.ff_fsize;
}

static long getdiskfree(const char drive){
	struct dfree df;
	long sz1, sz2;
	getdfree(drive -
			((drive >= 'a' && drive <= 'z') ? 'a': 'A') + 1, &df);
	if(df.df_sclus == 0xffff)
		return 0;
	sz1 = (long)df.df_avail;
	sz2 = (long)df.df_sclus * df.df_bsec;
	if((sz1 >> 8) * sz2 >= 2 * 1024L * 1024 * (1024 >> 8))
		return (long)(((unsigned long)~0) >> 1);
	return sz1 * sz2;
}

static void stringcopy(char *dest, const char *src, int n){
	int i;
	for(i = 1; i < n && *src; i++)
		*dest++ = *src++;
	*dest = 0;
}

static int isstring(char *str){
	int i, flag;
	char ch;
	flag = 0;
	for(i = 0;(ch = str[i]) != 0;i++){
		if(ch < '0' || ch > '9'){
			flag = 1;
			break;
		}
	}
	return flag;
}

int MakeNewFileName(
		const char *dir, const char *ext, long base, char *fname){
	int i, n;
	char fn[MAXPATH];
	long tmp, tmp2, *data, **stbl, *cp1, **cp2;
	struct ffblk fb;

	_ipdb_err_no = 0;
	fnmerge(fn, NULL, dir, "*", ext);
	fname[0] = 0;
	if(!(n = getfilenum(fn))){
		sprintf(fn, "%05ld", base);
		fnmerge(fname, NULL, NULL, fn, ext);
		return 0;
	}
	if((data = malloc((unsigned)sizeof(*data) * n)) == NULL)
		return (_ipdb_err_no = IPDB_MemoryAllocError);
	findfirst(fn, &fb, FA_ARCH | FA_RDONLY);
	cp1 = data;
	do{
		fnsplit(fb.ff_name, NULL, NULL, fn, NULL);
		if(!isstring(fn) && (tmp = atol(fn)) >= base)
			*(cp1++) = tmp;
	}while(!findnext(&fb));
	if((stbl = msort(data, sizeof(*data), n, 0)) == NULL){
		free(data);
		return _ipdb_err_no;
	}
	tmp = base;
	tmp2 = tmp + 1;
	for(i = 0, cp2 = stbl; i < n; i++, cp2++)
		if(tmp2 != **cp2){
			if(**cp2 != tmp)
				break;
			tmp++;
			tmp2 = **cp2;
		}
	free(stbl);
	free(data);
	sprintf(fn, "%05ld", tmp);
	fnmerge(fname, NULL, NULL, fn, ext);
	return 0;
}

static int chkhead(FILE *fp, int *ver){
	char head[sizeof(_headerstring)];
	int ver_no;

	_ipdb_err_no = 0;
	if(fread(head, sizeof(_headerstring) - 1, 1, fp) < 1)
		return (_ipdb_err_no = IPDB_ReadError);
	head[sizeof(_headerstring) - 1] = 0;
	if(!strcmp(head, _headerstring)){
			ver_no = 30;
	}else if(!strcmp(head, _headerstring20))
			ver_no = 20;
	else
		return (_ipdb_err_no = IPDB_IllegalHeader);
	if(ver != NULL)
		*ver = ver_no;
	return 0;
}

static int writehead(FILE *fp){
	_ipdb_err_no = 0;
	if(fwrite(_headerstring, sizeof(_headerstring) - 1, 1, fp) < 1)
		return (_ipdb_err_no = IPDB_WriteError);
	return 0;
}

static int copyfile(FILE *src, FILE *dest, long size){
	enum{ BufSize = 80 };
	char buf[BufSize];
	while(size > 0){
		if(size > BufSize){
			if(fread(buf, BufSize, 1, src) < 1)
				return (_ipdb_err_no = IPDB_ReadError);
			if(fwrite(buf, BufSize, 1, dest) < 1)
				return (_ipdb_err_no = IPDB_WriteError);
			size -= BufSize;
		}else{
			if(fread(buf, size, 1, src) < 1)
				return (_ipdb_err_no = IPDB_ReadError);
			if(fwrite(buf, size, 1, dest) < 1)
				return (_ipdb_err_no = IPDB_WriteError);
			size = 0;
		}
	}
	return (_ipdb_err_no = 0);
}

/* IPDB ̃t@C𓾂 */
int GetIPDBFileName(HIPDB hipdb, char fn_buf[], int buf_size){
	IPDB_INFO *ip;
	_ipdb_err_no = 0;
	fn_buf[0] = 0;
	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	stringcopy(fn_buf, ip->data_file, buf_size);
	return 0;
}

/* y[WN[Y */
int CloseIPDBPage(HIPDB hipdb){
	IPDB_INFO *ip;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL || ip->page.fp == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	fclose(ip->page.fp);
	ip->page.fp = NULL;
	return 0;
}

/* ݃I[vĂy[W̃CfNXԍ𓾂 */
int GetIPDBPageIndexNo(HIPDB hipdb){
	IPDB_INFO *ip;
	if((ip = getipdbinfo(hipdb)) == NULL ||
			ip->page.fp == NULL || !ip->page.acc_mode){
		_ipdb_err_no = IPDB_IllegalHandle;
		return NULL;
	}
	return ip->page.index_No;
}

/* ݃I[vĂy[Wԍ𓾂 */
unsigned GetIPDBPageNo(HIPDB hipdb){
	IPDB_INFO *ip;
	if((ip = getipdbinfo(hipdb)) == NULL ||
			ip->page.fp == NULL || !ip->page.acc_mode){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return ip->page.page_No;
}

/* y[WI[vĂ邩ǂׂ */
int IsIPDBPageOpened(HIPDB hipdb){
	IPDB_INFO *ip;
	if((ip = getipdbinfo(hipdb)) == NULL){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return (ip->page.fp != NULL && ip->page.acc_mode);
}

/* CfNXCfNXԍ𓾂 */
int GetIPDBIndexNo(HIPDB hipdb, const char index_name[]){
	int n = 1;
	IPDB_INFO *ip;
	INDEX_INFO *idx;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL){
		_ipdb_err_no = IPDB_IllegalHandle;
		return NULL;
	}
	idx = ip->index;
	while(idx != NULL){
		if(!strcmp(idx->name, index_name))
			break;
		n++;
		idx = idx->next;
	}
	return (idx != NULL) ? n: NULL;
}

static INDEX_INFO *getindexinfo(INDEX_INFO *idx, int index_No){
	if(index_No == NULL || index_No < 0)
		return NULL;
	while(idx != NULL && --index_No)
		idx = idx->next;
	return idx;
}

/* CfNXԍCfNX𓾂 */
int GetIPDBIndexName(HIPDB hipdb, int index_No, char name_buf[], int buf_size){
	IPDB_INFO *ip;
	INDEX_INFO *idx;
	_ipdb_err_no = 0;
	name_buf[0] = 0;
	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if((idx = getindexinfo(ip->index, index_No)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalValue);
	stringcopy(name_buf, idx->name, buf_size);
	return 0;
}

/* CfNX̐𓾂 */
int GetIPDBIndexCount(HIPDB hipdb){
	int n;
	IPDB_INFO *ip;
	INDEX_INFO *idx;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	idx = ip->index;
	for(n = 0; idx != NULL; n++, idx = idx->next);
	return n;
}

static int setpageinfo(IPDB_INFO *ip, int index_No, unsigned page_No){
	INDEX_INFO *idx;
	_ipdb_err_no = 0;
	if((idx = getindexinfo(
			ip->index, index_No)) == NULL || page_No >= idx->page_count)
		return (_ipdb_err_no = IPDB_IllegalValue);
	ip->page.index_No = index_No;
	ip->page.page_No = page_No;
	ip->page.index = idx;
	return 0;
}

static PAGE_INDEX *readpitbl(FILE *fp, int *ver){
	DWORD page_count;
	PAGE_INDEX *pi_tbl;

	_ipdb_err_no = 0;
	if(chkhead(fp, ver))
		return NULL;
	if(fread(&page_count, sizeof(page_count), 1, fp) < 1){
		_ipdb_err_no = IPDB_ReadError;
		return NULL;
	}
	if((pi_tbl = malloc(sizeof(PAGE_INDEX) * (page_count + 1))) == NULL){
		_ipdb_err_no = IPDB_MemoryAllocError;
		return NULL;
	}
	if(fread(&pi_tbl->index, sizeof(pi_tbl->index), 1, fp) < 1)
		_ipdb_err_no = IPDB_ReadError;
	else if(fseek(fp, pi_tbl->index, SEEK_SET))
		_ipdb_err_no = IPDB_SeekError;
	else if(fread(&pi_tbl[1],
			pi_tbl->size = page_count * sizeof(PAGE_INDEX), 1, fp) < 1)
		_ipdb_err_no = IPDB_ReadError;
	if(_ipdb_err_no){
		free(pi_tbl);
		return NULL;
	}
	return pi_tbl;
}

static long findbetterindex(PAGE_INDEX *pi_tbl,
		long num, long size, long *flag_size, long fsize){
	long i;
	long index = 0, min_size = 0x7fffffffL, tmp_size, prev_idx;
	PAGE_INDEX **stbl, **cp, *pi;

	_ipdb_err_no = 0;
	if((stbl = msort(pi_tbl, sizeof(*pi_tbl),
			num, (long)&(pi_tbl->index) - (long)pi_tbl)) == NULL)
		return NULL;
	*flag_size = 0;
	prev_idx = HeadDataSize;
	for(i = 0, cp = stbl; i < num; i++){
		pi = *(cp++);
		if(pi->size == 0)
			continue;
		if((tmp_size = pi->index - prev_idx) >= size && tmp_size < min_size){
			min_size = tmp_size;
			index = prev_idx;
		}
		*flag_size += pi->index - prev_idx;
		prev_idx = pi->index + pi->size;
	}
	if(index == 0){
		index = prev_idx;
		size = fsize - prev_idx - size;
		if(size > 0)
			*flag_size += size;
	}else
		*flag_size += fsize - prev_idx - size;
	free(stbl);
	return index;
}

/* y[WCgI[v */
int WriteOpenIPDBPage(
		HIPDB hipdb, int index_No, unsigned page_No, long size){
	IPDB_INFO *ip;
	PAGE_INDEX *pi_tbl, pi_tmp;
	long new_idx, page_count, fsize, flag_size;

	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL || ip->page.fp != NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if(size < 0)
		return (_ipdb_err_no = IPDB_IllegalValue);
	if(setpageinfo(ip, index_No, page_No))
		return _ipdb_err_no;
	fsize = getfilesize(ip->data_file);
	if((ip->page.fp = fopen(ip->data_file, "r+b")) == NULL)
		return (_ipdb_err_no = IPDB_FileOpenError);
	if((pi_tbl = readpitbl(ip->page.fp, NULL)) == NULL)
		goto err1;
	page_count = pi_tbl->size / sizeof(*pi_tbl);
	if((long)ip->page_count != page_count){
		_ipdb_err_no = IPDB_NoMatchMemory;
		goto err1;
	}
	page_No += ip->page.index->first_page;
	pi_tbl[page_No + 1].size = 0;
	if((new_idx = findbetterindex(
			pi_tbl, page_count + 1, size, &flag_size, fsize)) == NULL)
		goto err1;
	if(new_idx + size > fsize &&
			new_idx + size - fsize > getdiskfree(ip->data_file[0])){
		_ipdb_err_no = IPDB_WriteError;
		goto err1;
	}
	if(new_idx + size > fsize && size){
		CHAR tmp;
		fseek(ip->page.fp, new_idx + size - 1, SEEK_SET);
		if(fwrite(&tmp, sizeof(tmp), 1, ip->page.fp) < 1){
			_ipdb_err_no = IPDB_WriteError;
			goto err1;
		}
	}
	fseek(ip->page.fp, pi_tbl[0].index +
			sizeof(PAGE_INDEX) * (long)page_No, SEEK_SET);
	pi_tmp.index = new_idx;
	pi_tmp.size = size;
	if(fwrite(&pi_tmp, sizeof(pi_tmp), 1, ip->page.fp) < 1)
		_ipdb_err_no = IPDB_WriteError;
	else if(fseek(ip->page.fp, new_idx, SEEK_SET))
		_ipdb_err_no = IPDB_SeekError;
	err1:
	if(pi_tbl != NULL)
		free(pi_tbl);
	if(_ipdb_err_no){
		int tmp = _ipdb_err_no;
		CloseIPDBPage(hipdb);
		return (_ipdb_err_no = tmp);
	}
	ip->page.acc_mode = WriteAcc;
	ip->page.offset = new_idx;
	ip->page.size = size;
	ip->page.prev_acc = UpdateAcc;
	ip->flag_size = flag_size;
	return 0;
}

/* ݃I[vĂy[W̃TCY𓾂 */
long GetIPDBPageSize(HIPDB hipdb){
	IPDB_INFO *ip;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL ||
			ip->page.fp == NULL || !ip->page.acc_mode){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return ip->page.size;
}

int openpage(HIPDB hipdb, int index_No, unsigned page_No, int acc_mode){
	IPDB_INFO *ip;
	PAGE_INDEX pi_tmp;
	DWORD offset, page_count;

	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL || ip->page.fp != NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if(setpageinfo(ip, index_No, page_No))
		return _ipdb_err_no;
	if((ip->page.fp = fopen(
			ip->data_file, (acc_mode == ReadAcc) ? "rb": "r+b")) == NULL){
		_ipdb_err_no = IPDB_FileOpenError;
		goto err1;
	}
	if(chkhead(ip->page.fp, NULL))
		goto err1;
	page_No += ip->page.index->first_page;
	if(fread(&page_count, sizeof(page_count), 1, ip->page.fp) < 1 ||
			fread(&offset, sizeof(offset), 1, ip->page.fp) < 1)
		_ipdb_err_no = IPDB_ReadError;
	else if(ip->page_count != page_count)
		_ipdb_err_no = IPDB_NoMatchMemory;
	else if(fseek(ip->page.fp, offset +
			sizeof(PAGE_INDEX) * (long)page_No, SEEK_SET))
		_ipdb_err_no = IPDB_SeekError;
	else if(fread(&pi_tmp, sizeof(pi_tmp), 1, ip->page.fp) < 1)
		_ipdb_err_no = IPDB_ReadError;
	else{
		if(!pi_tmp.size)
			pi_tmp.index = 0;
		if(fseek(ip->page.fp, pi_tmp.index, SEEK_SET))
			_ipdb_err_no = IPDB_SeekError;
	}
	err1:
	if(_ipdb_err_no){
		int tmp = _ipdb_err_no;
		CloseIPDBPage(hipdb);
		return (_ipdb_err_no = tmp);
	}
	ip->page.acc_mode = acc_mode;
	ip->page.offset = pi_tmp.index;
	ip->page.size = pi_tmp.size;
	ip->page.prev_acc = UpdateAcc;
	return 0;
}

/* y[W[hI[v */
int ReadOpenIPDBPage(HIPDB hipdb, int index_No, unsigned page_No){
	return openpage(hipdb, index_No, page_No, ReadAcc);
}

/* y[W[hCgI[v */
int UpdateOpenIPDBPage(HIPDB hipdb, int index_No, unsigned page_No){
	return openpage(hipdb, index_No, page_No, UpdateAcc);
}

/* y[WV[N */
int SeekIPDBPage(HIPDB hipdb, long offset){
	IPDB_INFO *ip;
	if((ip = getipdbinfo(hipdb)) == NULL ||
			ip->page.fp == NULL || !ip->page.acc_mode)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if(offset >= ip->page.size || offset < 0)
		return (_ipdb_err_no = IPDB_IllegalValue);
	if(fseek(ip->page.fp, offset + ip->page.offset, SEEK_SET))
		return (_ipdb_err_no = IPDB_SeekError);
	ip->page.prev_acc = UpdateAcc;
	return (_ipdb_err_no = 0);
}

/* y[Wł݂̌̈ʒu𓾂 */
long GetIPDBPageOffset(HIPDB hipdb){
	IPDB_INFO *ip;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL ||
			ip->page.fp == NULL || !ip->page.acc_mode){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return ftell(ip->page.fp) - ip->page.offset;
}

/* y[WɃf[^ */
int WriteIPDBPage(HIPDB hipdb, void *data, long size){
	IPDB_INFO *ip;
	if((ip = getipdbinfo(hipdb)) == NULL || ip->page.fp == NULL ||
			(ip->page.acc_mode != WriteAcc && ip->page.acc_mode != UpdateAcc))
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if(size > ip->page.size -
			(ftell(ip->page.fp) - ip->page.offset) || size < 0)
		return (_ipdb_err_no = IPDB_IllegalValue);
	if(!size)
		return 0;
	if(ip->page.prev_acc == ReadAcc)
		fseek(ip->page.fp, 0, SEEK_CUR);
	ip->page.prev_acc = WriteAcc;
	if(fwrite(data, size, 1, ip->page.fp) < 1)
		return (_ipdb_err_no = IPDB_WriteError);
	return (_ipdb_err_no = 0);
}

/* y[Wf[^ǂݍ */
int ReadIPDBPage(HIPDB hipdb, void *data, long size){
	IPDB_INFO *ip;
	if((ip = getipdbinfo(hipdb)) == NULL || ip->page.fp == NULL ||
			(ip->page.acc_mode != ReadAcc && ip->page.acc_mode != UpdateAcc))
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if(size > ip->page.size -
			(ftell(ip->page.fp) - ip->page.offset) || size < 0)
		return (_ipdb_err_no = IPDB_IllegalValue);
	if(!size)
		return 0;
	if(ip->page.prev_acc == WriteAcc)
		fseek(ip->page.fp, 0, SEEK_CUR);
	ip->page.prev_acc = ReadAcc;
	if(fread(data, size, 1, ip->page.fp) < 1)
		return (_ipdb_err_no = IPDB_ReadError);
	return (_ipdb_err_no = 0);
}

static void freeindexlist(INDEX_INFO *idx){
	void *tmp;
	while(idx != NULL){
		tmp = idx->next;
		free(idx);
		idx = tmp;
	}
}

/* IPDB  */
int ReleaseIPDB(HIPDB hipdb){
	IPDB_INFO *ip, **tmp;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL){
		return (_ipdb_err_no = IPDB_IllegalHandle);
	}
	if(ip->page.fp != NULL)
		CloseIPDBPage(hipdb);
	freeindexlist(ip->index);
	tmp = &_ipdb_info_chain;
	while((ip = *tmp) != NULL){
		if(ip->handle_info.handle == hipdb)
			break;
		tmp = &(ip->chain);
	}
	if(ip != NULL)
		*tmp = ip->chain;
	removehandleinfo(hipdb, HT_IPDB);
	return _ipdb_err_no;
}

static IPDB_INFO *fntoipdb(const char *fn){
	IPDB_INFO *ip = _ipdb_info_chain;
	while(ip != NULL){
		if(!strcmp(ip->data_file, fn))
			break;
		ip = ip->chain;
	}
	return ip;
}

static INDEX_INFO *addindex(
		IPDB_INFO *ip, const char index_name[], unsigned page_count){
	INDEX_INFO *tmp, **idx;
	unsigned page_pos;

	_ipdb_err_no = 0;
	if((long)ip->page_count + page_count > MaxPageCount){
		_ipdb_err_no = IPDB_IllegalValue;
		return NULL;
	}
	if(!index_name[0]){
		_ipdb_err_no = IPDB_IllegalName;
		return NULL;
	}
	page_pos = 1;
	idx = &ip->index;
	while((tmp = *idx) != NULL){
		if(!strncmp(tmp->name, index_name, sizeof(tmp->name) - 1)){
			_ipdb_err_no = IPDB_AlreadyExist;
			return NULL;
		}
		page_pos += tmp->page_count;
		idx = &tmp->next;
	}
	if((tmp = malloc(sizeof(*tmp))) == NULL){
		_ipdb_err_no = IPDB_MemoryAllocError;
		return NULL;
	}
	stringcopy(tmp->name, index_name, sizeof(tmp->name));
	tmp->first_page = page_pos;
	tmp->page_count = page_count;
	tmp->next = NULL;
	ip->page_count = page_pos + page_count;
	*idx = tmp;
	return tmp;
}

static INDEX_INFO *delindex(IPDB_INFO *ip, const char index_name[]){
	INDEX_INFO *tmp, **idx;

	if(ip->index == NULL)
		return NULL;
	idx = &ip->index->next;
	while((tmp = *idx) != NULL){
		if(!strcmp(tmp->name, index_name))
			break;
		idx = &tmp->next;
	}
	if(tmp != NULL)
		*idx = tmp->next;
	return tmp;
}

static long getindexinfosize(INDEX_INFO *idx){
	long size = 1;

	while(idx != NULL){
		size += strlen(idx->name) + 2 * 2 + 1;
		idx = idx->next;
	}
	return size;
}

static int writeindexinfo(FILE *fp, INDEX_INFO *idx){
	CHAR len;
	while(idx != NULL){
		len = strlen(idx->name);
		fwrite(&len, sizeof(len), 1, fp);
		fwrite(idx->name, len, 1, fp);
		fwrite(&idx->first_page, sizeof(idx->first_page), 1, fp);
		fwrite(&idx->page_count, sizeof(idx->page_count), 1, fp);
		idx = idx->next;
	}
	len = 0;
	if(fwrite(&len, sizeof(len), 1, fp) < 1)
		return IPDB_WriteError;
	return 0;
}

static INDEX_INFO *readindexinfo(FILE *fp){
	BYTE len;
	INDEX_INFO *idx, **tail, *tmp;
	_ipdb_err_no = 0;
	idx = NULL;
	tail = &idx;
	for(;;){
		if(fread(&len, sizeof(len), 1, fp) < 1){
			_ipdb_err_no = IPDB_ReadError;
			break;
		}
		if(!len)
			break;
		if(len >= sizeof(tmp->name)){
			_ipdb_err_no = IPDB_IllegalFile;
			break;
		}
		if((tmp = malloc(sizeof(*tmp))) == NULL){
			_ipdb_err_no = IPDB_ReadError;
			break;
		}
		fread(tmp->name, len, 1, fp);
		tmp->name[len] = 0;
		fread(&tmp->first_page, sizeof(tmp->first_page), 1, fp);
		fread(&tmp->page_count, sizeof(tmp->page_count), 1, fp);
		tmp->next = NULL;
		*tail = tmp;
		tail = &tmp->next;
	}
	if(_ipdb_err_no && idx != NULL){
		freeindexlist(idx);
		idx = NULL;
	}
	return idx;
}

static int createipdbfile(IPDB_INFO *ip){
	FILE *fp;
	LONG tmp;
	PAGE_INDEX buf;

	_ipdb_err_no = 0;
	if((fp = fopen(ip->data_file, "wb")) == NULL)
		return (_ipdb_err_no = IPDB_FileOpenError);
	if(!writehead(fp)){
		tmp = 1;
		fwrite(&tmp, sizeof(tmp), 1, fp);
		tmp = ftell(fp) + sizeof(tmp);
		if(fwrite(&tmp, sizeof(tmp), 1, fp) < 1)
			_ipdb_err_no = IPDB_WriteError;
		else{
			buf.index = tmp + sizeof(buf);
			buf.size = getindexinfosize(ip->index);
			fwrite(&buf, sizeof(buf), 1, fp);
			if(writeindexinfo(fp, ip->index))
				_ipdb_err_no = IPDB_WriteError;
		}
	}
	fclose(fp);
	if(_ipdb_err_no)
		remove(ip->data_file);
	return _ipdb_err_no;
}

static char tolarge(char ch){
	if(ch >= 'a' && ch <= 'z')
		return ch + 'A' - 'a';
	return ch;
}

static void stol(char *str){
	while(*str)
		*str++ = tolarge(*str);
}

static int getfullpath(const char fn[], char buf[]){
	char dr[MAXDRIVE], di[MAXDIR], fi[MAXFILE], ex[MAXEXT];
	char cur[MAXDRIVE + MAXDIR - 1];
	int len;

	_ipdb_err_no = 0;
	fnsplit(fn, dr, di, fi, ex);
	fnmerge(cur, dr, di, NULL, NULL);
	getcwd(buf, MAXPATH);
	if((len = strlen(cur)) > 0)
		if(cur[len - 1] == '\\' && len > 1)
			cur[len - 1] = 0;
	if(di[0])
		if(chdir(cur)){
			buf[0] = 0;
			return (_ipdb_err_no = IPDB_IllegalName);
		}
	if(dr[0])
		setdisk(tolarge(dr[0]) - 'A');
	getcwd(cur, sizeof(cur));
	chdir(buf);
	setdisk(buf[0] - 'A');
	stol(fi);
	stol(ex);
	fnmerge(buf, NULL, cur, fi, ex);
	return 0;
}

static IPDB_INFO *createipdb(const char *data_file){
	IPDB_INFO *ip;
	_ipdb_err_no = 0;
	if((ip = makehandleinfo(HT_IPDB, sizeof(IPDB_INFO))) == NULL){
		_ipdb_err_no = (gethandlecount() ==
				MaxHandleCount) ? IPDB_HandleOver: IPDB_MemoryAllocError;
		return NULL;
	}
	stringcopy(ip->data_file, data_file, sizeof(ip->data_file));
	getfullpath(ip->data_file, ip->data_file);
//	if(fntoipdb(ip->data_file) != NULL)
//		_ipdb_err_no = IPDB_AlreadyExist;
	if(_ipdb_err_no){
		removehandleinfo(ip->handle_info.handle, HT_IPDB);
		return NULL;
	}
	ip->index = NULL;
	ip->chain = _ipdb_info_chain;
	_ipdb_info_chain = ip;
	ip->page_count = 0;
	ip->page.fp = NULL;
	ip->flag_size = 0;
	return ip;
}

/* p^[f[^x[XVK쐬 */
HIPDB CreateIPDB(const char data_file[]){
	IPDB_INFO *ip;

	_ipdb_err_no = 0;
	if(getfilenum(data_file)){
		_ipdb_err_no = IPDB_AlreadyExist;
		return NULL;
	}
	if((ip = createipdb(data_file)) == NULL)
		return NULL;
	if(createipdbfile(ip)){
		int tmp = _ipdb_err_no;
		ReleaseIPDB(ip->handle_info.handle);
		_ipdb_err_no = tmp;
		return NULL;
	}
	return ip->handle_info.handle;
}

static unsigned sumidxpagecount(INDEX_INFO *idx){
	unsigned page_count = 1;
	while(idx != NULL){
		page_count += idx->page_count;
		idx = idx->next;
	}
	return page_count;
}

static long getflagsize(PAGE_INDEX *pi_tbl, long num, long fsize){
	long i;
	long prev_idx, size = 0;
	PAGE_INDEX **stbl, **cp, *pi;

	_ipdb_err_no = 0;
	if((stbl = msort(pi_tbl, sizeof(*pi_tbl),
			num, (long)&(pi_tbl->index) - (long)pi_tbl)) == NULL)
		return 0;
	prev_idx = HeadDataSize;
	for(i = 0, cp = stbl; i < num; i++, cp++){
		pi = *cp;
		if(pi->size == 0)
			continue;
		size += pi->index - prev_idx;
		prev_idx = pi->index + pi->size;
	}
	size += fsize - prev_idx;
	free(stbl);
	return size;
}

/* IPDB ǂݍ */
HIPDB LoadIPDB(const char data_file[]){
	FILE *fp;
	IPDB_INFO *ip = NULL;
	int ver;
	long page_count, flag_size, fsize;
	PAGE_INDEX *pi_tbl;

	_ipdb_err_no = 0;
	{
		char buf[MAXPATH];
		getfullpath(data_file, buf);
		if((ip = fntoipdb(buf)) != NULL)
			return ip->handle_info.handle;
	}
	fsize = getfilesize(data_file);
	if((fp = fopen(data_file, "rb")) == NULL){
		_ipdb_err_no = IPDB_FileOpenError;
		return NULL;
	}
	if((pi_tbl = readpitbl(fp, &ver)) == NULL)
		goto err1;
	page_count = pi_tbl[0].size / sizeof(*pi_tbl);
	flag_size = getflagsize(pi_tbl, page_count + 1, fsize);
	if(!_ipdb_err_no && (ip = createipdb(data_file)) != NULL)
		if(ver == 30){
			if(!fseek(fp, pi_tbl[1].index, SEEK_SET))
				ip->index = readindexinfo(fp);
			else
				_ipdb_err_no = IPDB_SeekError;
		}else{
			if(addindex(ip, _aif_name, 1) == NULL ||
					addindex(ip, _set_ai_name, 1) == NULL ||
					addindex(ip, _opdb_name, page_count - 3) == NULL);
		}
	free(pi_tbl);
	if(!_ipdb_err_no){
		if((long)sumidxpagecount(ip->index) == page_count){
			ip->page_count = page_count;
			ip->flag_size = flag_size;
		}else
			_ipdb_err_no = IPDB_IllegalFile;
	}
err1:
	fclose(fp);
	if(_ipdb_err_no){
		if(ip != NULL){
			int tmp = _ipdb_err_no;
			ReleaseIPDB(ip->handle_info.handle);
			_ipdb_err_no = tmp;
		}
		return NULL;
	}
	return ip->handle_info.handle;
}

/* y[W𓾂 */
unsigned GetIPDBIndexPageCount(HIPDB hipdb, int index_No){
	IPDB_INFO *ip;
	INDEX_INFO *idx;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	if((idx = getindexinfo(ip->index, index_No)) == NULL){
		_ipdb_err_no = IPDB_IllegalValue;
		return 0;
	}
	return idx->page_count;
}

static void addidxpageno(INDEX_INFO *idx, long page_count){
	while(idx != NULL){
		idx->first_page += (WORD)page_count;
		idx = idx->next;
	}
}

static int addpage(IPDB_INFO *ip, unsigned page_pos, long page_count){
	long i;
	FILE *fp;
	LONG now_page_count, flag_size, fsize;
	PAGE_INDEX *pi_tbl, *pi_cur, pi;

	_ipdb_err_no = 0;
	fsize = getfilesize(ip->data_file);
	if((fp = fopen(ip->data_file, "r+b")) == NULL)
		return (_ipdb_err_no = IPDB_FileOpenError);
	if((pi_tbl = readpitbl(fp, NULL)) == NULL)
		goto err1;
	now_page_count = pi_tbl[0].size / sizeof(PAGE_INDEX);
	if(now_page_count != (LONG)(ip->page_count - page_count)){
		_ipdb_err_no = IPDB_NoMatchMemory;
		goto err2;
	}
	pi_tbl[0].size = 0;
	if((pi_tbl[0].index = findbetterindex(pi_tbl, now_page_count + 1,
			pi.size = sizeof(PAGE_INDEX) * (now_page_count + page_count),
			&flag_size, fsize)) == NULL)
		goto err2;
	if(pi_tbl[0].index + pi.size > fsize)
		fsize = pi_tbl[0].index + pi.size;
	pi_tbl[0].size = pi.size;
	pi_tbl[1].size = 0;
	pi_cur = &pi_tbl[page_pos + 1];
	for(i = 0; i < -page_count; i++, pi_cur++)
		pi_cur->size = 0;
	if((pi_tbl[1].index = findbetterindex(pi_tbl, now_page_count + 1,
			pi.size = getindexinfosize(ip->index),
			&flag_size, fsize)) == NULL)
		goto err2;
	pi_tbl[1].size = pi.size;
	if(pi_tbl[1].index + pi.size > fsize &&
			pi_tbl[1].index + pi.size - fsize > getdiskfree(ip->data_file[0])){
		_ipdb_err_no = IPDB_WriteError;
		goto err2;
	}
	if(fseek(fp, pi_tbl[0].index, SEEK_SET)){
		_ipdb_err_no = IPDB_SeekError;
		goto err2;
	}
	if(fwrite(&pi_tbl[1], sizeof(PAGE_INDEX) * page_pos, 1, fp) < 1){
		_ipdb_err_no = IPDB_WriteError;
		goto err2;
	}
	pi.index = pi.size = 0;
	for(i = 0; i < page_count; i++)
		if(fwrite(&pi, sizeof(pi), 1, fp) < 1){
			_ipdb_err_no = IPDB_WriteError;
			goto err2;
		}
	if((pi.size = now_page_count -
			page_pos + ((page_count < 0) ? page_count: 0)) > 0)
		if(fwrite(&pi_tbl[page_pos - ((page_count < 0) ? page_count: 0) + 1],
				sizeof(PAGE_INDEX) * pi.size, 1, fp) < 1){
			_ipdb_err_no = IPDB_WriteError;
			goto err2;
		}
	now_page_count += page_count;
	if(fseek(fp, sizeof(_headerstring) - 1, SEEK_SET))
		_ipdb_err_no = IPDB_SeekError;
	else if(fwrite(&now_page_count, sizeof(now_page_count), 1, fp) < 1 ||
			fwrite(&pi_tbl[0].index, sizeof(pi_tbl[0].index), 1, fp) < 1)
		_ipdb_err_no = IPDB_WriteError;
	else if(fseek(fp, pi_tbl[1].index, SEEK_SET))
		_ipdb_err_no = IPDB_SeekError;
	else if(writeindexinfo(fp, ip->index))
		_ipdb_err_no = IPDB_WriteError;
	err2:
	free(pi_tbl);
	err1:
	fclose(fp);
	if(!_ipdb_err_no)
		ip->flag_size = flag_size;
	return _ipdb_err_no;
}

/* CfNX */
int MakeIPDBIndex(HIPDB hipdb, const char index_name[], unsigned page_count){
	IPDB_INFO *ip;
	INDEX_INFO *idx;

	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if((idx = addindex(ip, index_name, page_count)) == NULL)
		return _ipdb_err_no;
	if(addpage(ip, idx->first_page, page_count)){
		char buf[sizeof(idx->name)];
		stringcopy(buf, index_name, sizeof(buf));
		delindex(ip, buf);
		free(idx);
		return _ipdb_err_no;
	}
	return (_ipdb_err_no = 0);
}

/* CfNX폜 */
int RemoveIPDBIndex(HIPDB hipdb, const char index_name[]){
	int index_No;
	IPDB_INFO *ip;
	INDEX_INFO *idx;
	unsigned first_page, page_count;

	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if((idx = delindex(ip, index_name)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalName);
	first_page = idx->first_page;
	page_count = idx->page_count;
	if((index_No = GetIPDBIndexNo(hipdb, index_name)) == ip->page.index_No)
		ip->page.acc_mode = 0;
	else if(ip->page.index_No > index_No)
		ip->page.index_No--;
	addidxpageno(idx->next, -(long)page_count);
	ip->page_count -= page_count;
	free(idx);
	if(addpage(ip, first_page, -(long)page_count))
		return _ipdb_err_no;
	return (_ipdb_err_no = 0);
}

/* y[Wǉ */
int AddIPDBIndexPages(HIPDB hipdb,
		int index_No, unsigned page_pos, unsigned page_count){
	IPDB_INFO *ip;
	INDEX_INFO *idx;

	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if((idx = getindexinfo(ip->index,
			index_No)) == NULL || page_pos > idx->page_count)
		return (_ipdb_err_no = IPDB_IllegalValue);
	if(ip->page.index_No == index_No && ip->page.page_No >= page_pos)
		ip->page.page_No += page_count;
	addidxpageno(idx->next, page_count);
	idx->page_count += page_count;
	ip->page_count += page_count;
	if(addpage(ip, idx->first_page + page_pos, page_count))
		return _ipdb_err_no;
	return (_ipdb_err_no = 0);
}

/* y[W폜 */
int DeleteIPDBIndexPages(HIPDB hipdb,
		int index_No, unsigned page_pos, unsigned page_count){
	IPDB_INFO *ip;
	INDEX_INFO *idx;

	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if((idx = getindexinfo(ip->index, index_No)) == NULL ||
			page_pos >= idx->page_count ||
			idx->page_count - ((long)page_pos + page_count) < 0)
		return (_ipdb_err_no = IPDB_IllegalValue);
	if(ip->page.index_No == index_No && ip->page.page_No >= page_pos)
		if(ip->page.page_No >= (long)page_pos + page_count)
			ip->page.page_No -= page_count;
		else
			ip->page.acc_mode = 0;
	addidxpageno(idx->next, -(long)page_count);
	idx->page_count -= page_count;
	ip->page_count -= page_count;
	if(addpage(ip, idx->first_page + page_pos, -(long)page_count))
		return _ipdb_err_no;
	return (_ipdb_err_no = 0);
}

/* IPDB ̖gp̈̃TCY𓾂 */
long GetIPDBFlagSize(HIPDB hipdb){
	IPDB_INFO *ip;
	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL){
		_ipdb_err_no = IPDB_IllegalHandle;
		return 0;
	}
	return ip->flag_size;
}

/* IPDB ̖gp̈폜 */
int ArrangeIPDB(HIPDB hipdb){
	unsigned i, page_No = 0;
	PAGE_INDEX *page_tbl, *pt;
	LONG page_count, tbl_idx, offset, index, fsize;
	char ufname[MAXPATH];
	FILE *fp, *wfp;
	IPDB_INFO *ip;

	_ipdb_err_no = 0;
	if((ip = getipdbinfo(hipdb)) == NULL)
		return (_ipdb_err_no = IPDB_IllegalHandle);
	if(!ip->flag_size)
		return 0;
	if(ip->page.fp != NULL){
		if(ip->page.acc_mode)
			page_No = ip->page.index->first_page + ip->page.page_No;
		offset = ftell(ip->page.fp) - ip->page.offset;
		fclose(ip->page.fp);
	}
	{
		char dr[MAXDRIVE], di[MAXDIR];
		fnsplit(ip->data_file, dr, di, NULL, NULL);
		fnmerge(ufname, dr, di, "t)m(p()(.)$(", NULL);
	}
	fsize = getfilesize(ip->data_file);
	if((fp = fopen(ip->data_file, "rb")) == NULL)
		return (_ipdb_err_no = IPDB_FileOpenError);
	if((wfp = fopen(ufname, "w+b")) == NULL){
		fclose(fp);
		return (_ipdb_err_no = IPDB_FileOpenError);
	}
	if(fsize - ip->flag_size > getdiskfree(ip->data_file[0])){
		_ipdb_err_no = IPDB_WriteError;
		goto err1;
	}
	if(chkhead(fp, NULL))
		goto err1;
	if(fread(&page_count, sizeof(page_count), 1, fp) < 1 ||
			fread(&tbl_idx, sizeof(tbl_idx), 1, fp) < 1){
		_ipdb_err_no = IPDB_ReadError;
		goto err1;
	}
	if(fseek(fp, tbl_idx, SEEK_SET)){
		_ipdb_err_no = IPDB_SeekError;
		goto err1;
	}
	if(writehead(wfp))
		goto err1;
	tbl_idx = HeadDataSize;
	if(fwrite(&page_count, sizeof(page_count), 1, wfp) < 1 ||
			fwrite(&tbl_idx, sizeof(tbl_idx), 1, wfp) < 1){
		_ipdb_err_no = IPDB_WriteError;
		goto err1;
	}
	if((page_tbl = malloc(sizeof(*page_tbl) * page_count)) == NULL){
		_ipdb_err_no = IPDB_MemoryAllocError;
		goto err1;
	}
	if(fread(page_tbl, sizeof(*page_tbl) * page_count, 1, fp) < 1){
		_ipdb_err_no = IPDB_ReadError;
		goto err2;
	}
	if(fwrite(page_tbl, sizeof(*page_tbl) * page_count, 1, wfp) < 1){
		_ipdb_err_no = IPDB_WriteError;
		goto err2;
	}
	tbl_idx += sizeof(*page_tbl) * page_count;
	for(i = 0, pt = page_tbl; (LONG)i < page_count; i++, pt++){
		if(!pt->size){
			if(!i){
				pt->size = getindexinfosize(ip->index);
				if(writeindexinfo(wfp, ip->index)){
					_ipdb_err_no = IPDB_WriteError;
					goto err2;
				}
				pt->index = tbl_idx;
				tbl_idx += pt->size;
				continue;
			}
			pt->index = 0;
			continue;
		}
		if(fseek(fp, pt->index, SEEK_SET)){
			_ipdb_err_no = IPDB_SeekError;
			goto err2;
		}
		pt->index = tbl_idx;
		tbl_idx += pt->size;
		if(copyfile(fp, wfp, pt->size))
			goto err2;
	}
	if(fseek(wfp, HeadDataSize, SEEK_SET)){
		_ipdb_err_no = IPDB_SeekError;
		goto err2;
	}
	if(fwrite(page_tbl, sizeof(*page_tbl) * page_count, 1, wfp) < 1)
		_ipdb_err_no = IPDB_WriteError;
	index = page_tbl[page_No].index;
	err2:
	free(page_tbl);
	err1:
	fclose(wfp);
	fclose(fp);
	if(!_ipdb_err_no){
		remove(ip->data_file);
		rename(ufname, ip->data_file);
		ip->flag_size = 0;
	}else
		remove(ufname);
	if(ip->page.fp != NULL){
		if(!_ipdb_err_no)
			ip->page.offset = index;
		if((ip->page.fp = fopen(ip->data_file, "r+b")) != NULL){
			if(!fseek(ip->page.fp, ip->page.offset + offset, SEEK_SET)){
				ip->page.prev_acc = UpdateAcc;
			}else if(!_ipdb_err_no)
				_ipdb_err_no = IPDB_SeekError;
		}else if(!_ipdb_err_no)
			_ipdb_err_no = IPDB_FileOpenError;
	}
	return _ipdb_err_no;
}

/* ڂG[bZ[W𓾂 */
int GetIPDBError(void){
	return _ipdb_err_no;
}
