#include <stdio.h>
#include <stdlib.h>
#include <jctype.h>
#include <dos.h>
#include <dir.h>
#include "mdbio2.h"

static OPDB_XY _frame_margin = {
	10000, 10000
};

void setframemargin(OPDB_XY *sz){
	_frame_margin = *sz;
}

void getframemargin(OPDB_XY *sz){
	*sz = _frame_margin;
}

static void freecl(CHECK_LIST *cl){
	if(cl == NULL)
		return;
	if((cl->flag & MCF_Checked) &&
			(cl->flag & MCF_Illegal) && cl->err_kind == EK_Comment){
		if(cl->data.comment != NULL)
			free(cl->data.comment);
	}
	free(cl);
}

void freechecklist(CHECK_LIST **cl, int count){
	int i;
	for(i = 0; i < count; i++)
		freecl(cl[i]);
	free(cl);
}

static CHECK_LIST *allocchecklist(void){
	return calloc(sizeof(CHECK_LIST), 1);
}

CHECK_LIST **createchecklist(int size){
	int i;
	CHECK_LIST **cl;
	if((cl = calloc(sizeof(*cl), size)) == NULL)
		return NULL;
	for(i = 0; i < size; i++)
		if((cl[i] = allocchecklist()) == NULL){
			freechecklist(cl, i);
			return NULL;
		}
	return cl;
}

static int readchecklist(FILE *fp, CHECK_LIST *cl){
	int err = 0;
	unsigned char len;
	err |= fread(&cl->flag, sizeof(cl->flag), 1, fp) < 1;
	err |= fread(&cl->err_kind, sizeof(cl->err_kind), 1, fp) < 1;
	err |= fread(&cl->score, sizeof(cl->score), 1, fp) < 1;
	if((cl->flag & MCF_Checked) && (cl->flag & MCF_Illegal)){
		switch(cl->err_kind){
		case EK_Other:
			err |= fread(&cl->data.code, sizeof(cl->data.code), 1, fp) < 1;
			break;
		case EK_Comment:
			err |= fread(&len, sizeof(len), 1, fp) < 1;
			if((cl->data.comment = malloc(sizeof(char) * (len + 1))) == NULL){
				err = 1;
				break;
			}
			if(len)
				err |= fread(cl->data.comment, sizeof(char) * len, 1, fp) < 1;
			cl->data.comment[len] = 0;
			break;
		}
	}
	return err;
}

int loadchecklist(const char *fn, CHECK_LIST **cl, int count){
	int err = 1;
	int i;
	FILE *fp;
	if((fp = fopen(fn, "rb")) == NULL)
		return 1;
	for(i = 0; i < count; i++)
		if(readchecklist(fp, cl[i]))
			goto err1;
	err = 0;
err1:
	fclose(fp);
	return err;
}

static int writechecklist(FILE *fp, CHECK_LIST *cl){
	int err = 0;
	unsigned char len;
	err |= fwrite(&cl->flag, sizeof(cl->flag), 1, fp) < 1;
	err |= fwrite(&cl->err_kind, sizeof(cl->err_kind), 1, fp) < 1;
	err |= fwrite(&cl->score, sizeof(cl->score), 1, fp) < 1;
	if((cl->flag & MCF_Checked) && (cl->flag & MCF_Illegal)){
		switch(cl->err_kind){
		case EK_Other:
			err |= fwrite(&cl->data.code, sizeof(cl->data.code), 1, fp) < 1;
			break;
		case EK_Comment:
			if(cl->data.comment == NULL)
				len = 0;
			else
				len = strlen(cl->data.comment);
			err |= fwrite(&len, sizeof(len), 1, fp) < 1;
			err |= fwrite(cl->data.comment, sizeof(char) * len, 1, fp) < 1;
			break;
		}
	}
	return err;
}

int savechecklist(const char *fn, CHECK_LIST **cl, int count){
	int err = 1;
	int i;
	FILE *fp;
	if((fp = fopen(fn, "wb")) == NULL)
		return 1;
	for(i = 0; i < count; i++)
		if(writechecklist(fp, cl[i]))
			goto err1;
	err = 0;
err1:
	fclose(fp);
	return err;
}

static long gettextsize(const char fn[]){
	int ch;
	long size = 0;
	unsigned moji;
	FILE *fp;
	if((fp = fopen(fn, "rt")) == NULL)
		return 0;
	while((ch = fgetc(fp)) != EOF){
		moji = iskanji(ch) ? ((ch << 8) | fgetc(fp)): ch;
		if(moji == L'@' || moji <= ' ')
			continue;
		size++;
	}
	fclose(fp);
	return size;
}

WORD *loadtext(const char *fname){
	int ch, err;
	struct ffblk fb;
	FILE *fp;
	WORD moji, *text, *cur;
	long size;
	if(findfirst(fname, &fb, FA_ARCH | FA_RDONLY))
		return NULL;
	size = gettextsize(fname);
	if((cur = text = malloc(sizeof(WORD) * (size + 1))) == NULL)
		return NULL;
	err = (fp = fopen(fname, "rt")) == NULL;
	while(!err && (ch = fgetc(fp)) != EOF){
		if(iskanji(moji = (unsigned)ch)){
			err = (ch = fgetc(fp)) == EOF;
			moji = (moji << 8) | ch;
		}
		if(moji > ' ' && moji != L'@')
			*cur++ = moji;
	}
	*cur = 0;
	if(fp != NULL) fclose(fp);
	if(err){
		if(text != NULL) free(text);
		text = NULL;
	}
	return text;
}

int textlen(WORD *text){
	int i;
	for(i = 0; *text++ != 0; i++);
	return i;
}

static void freestlist(STROKE_LIST *st){
	STROKE_LIST *next;
	while(st != NULL){
		next = st->next;
		free(st->pt);
		free(st);
		st = next;
	}
}

void initmp(MOJI_PATTERN *mp){
	freestlist(mp->head);
	mp->head = NULL;
	mp->tail = &(mp->head);
}

int addmpstroke(MOJI_PATTERN *mp, OPDB_XY *pbuf, unsigned pt_count){
	STROKE_LIST *st;
	if((st = malloc(sizeof(*st))) == NULL)
		return 1;
	st->pt = pbuf;
	st->count = pt_count;
	st->next = NULL;
	*mp->tail = st;
	mp->tail = &st->next;
	return 0;
}

void initmparray(MOJI_PATTERN *mp, int frame_count){
	int i;
	for(i = 0; i < frame_count; i++)
		initmp(mp++);
}

MOJI_PATTERN *makemparray(int frame_count){
	int i;
	MOJI_PATTERN *mp, *cur;
	if((mp = cur = malloc(frame_count * sizeof(*mp))) == NULL)
		return NULL;
	for(i = 0; i < frame_count; i++, cur++){
		cur->head = NULL;
		cur->tail = &cur->head;
	}
	return mp;
}

void freemparray(MOJI_PATTERN *mp, int frame_count){
	initmparray(mp, frame_count);
	free(mp);
}

int getnearframeno(FRAME *frame, OPDB_XY *pt){
	int m, n, limit;
	n = 0;
	limit = frame->start.y + (frame->size.y - frame->step.y) / 2;
	if(pt->y >= limit){
		n = (pt->y - limit) / frame->step.y;
		if(n >= frame->count.y)
			n = frame->count.y - 1;
	}
	m = 0;
	limit = frame->start.x + (frame->size.x - frame->step.x) / 2;
	if(pt->x >= limit){
		m = (pt->x - limit) / frame->step.x;
		if(m >= frame->count.x)
			m = frame->count.x - 1;
	}
	if(abs(frame->step.x * m + (frame->size.x >> 1) +
			frame->start.x - pt->x) >
			frame->size.x + _frame_margin.x ||
			abs(frame->step.y * n + (frame->size.y >> 1) +
			frame->start.y - pt->y) > frame->size.y +
			_frame_margin.y)
		return -1;
	return n * frame->count.x + m;
}

int savemparray(HOPDB hOpdb, int page_No, MOJI_PATTERN *mp, int frame_count){
	int i, tag_No_ofs, item_No, err;
	unsigned stroke_No = 0;
	HPAGEPAT hpp;
	STROKE_LIST *st;
	HAIFMT stroke_haif;
	stroke_haif = GetOPDBAddInfoFormat(hOpdb, AIF_Stroke);
	if((item_No = GetAddInfoItemNo(stroke_haif, "tag No")) != NULL)
		tag_No_ofs = GetAddInfoItemOffset(stroke_haif, item_No);
	err = (hpp = CreateOPDBPagePattern(hOpdb, page_No)) == NULL;
	for(i = 0; i < frame_count && !err; i++, mp++){
		st = mp->head;
		while(st != NULL){
			err = AddOPDBStrokeXY(hpp, stroke_No, st->pt, NULL, st->count);
			if(err)
				break;
			if(item_No != NULL)
				SetOPDBStrokeAddInfo(
						hpp, stroke_No++, tag_No_ofs, &i, sizeof(i));
			st = st->next;
		}
	}
	ReleaseAddInfoFormat(stroke_haif);
	if(!err) err = SaveOPDBPagePattern(hpp);
	if(hpp) ReleaseOPDBPagePattern(hpp);
	return err;
}

void getframeorgpos(FRAME *frame, int n, OPDB_XY *org){
	org->x = (n % frame->count.x) * frame->step.x + frame->start.x;
	org->y = (n / frame->count.x) * frame->step.y + frame->start.y;
}

void getstrokerect(OPDB_RECT *rc, OPDB_XY *pt, unsigned pt_count){
	int i;
	rc->left = rc->right = pt->x;
	rc->top = rc->bottom = pt->y;
	for(i = 1, pt++; i < pt_count; i++, pt++){
		if(rc->left > pt->x)
			rc->left = pt->x;
		else if(rc->right < pt->x)
			rc->right = pt->x;
		if(rc->top > pt->y)
			rc->top = pt->y;
		else if(rc->bottom < pt->y)
			rc->bottom = pt->y;
	}
}

void convscale(OPDB_XY *src, OPDB_XY *dst, OPDB_XY *pt){
	pt->x = (int)((long)pt->x * dst->x * 2 / src->x + 1 >> 1);
	pt->y = (int)((long)pt->y * dst->y * 2 / src->y + 1 >> 1);
}

static int addstroke(
		HPAGEPAT hpp, unsigned stroke_No, MOJI_PATTERN mp[],
		FRAME *frame, OPDB_XY *in_reso, OPDB_XY *disp_reso){
	unsigned pt_count;
	int n;
	OPDB_XY pt, *op;
	OPDB_RECT rc;
	pt_count = GetOPDBPointCount(hpp, stroke_No);
	if((op = malloc(sizeof(*op) * pt_count)) == NULL)
		return OPDB_MemoryAllocError;
	GetOPDBStrokeXY(hpp, stroke_No, op, NULL, pt_count);
	getstrokerect(&rc, op, pt_count);
	pt.x = rc.left + rc.right >> 1;
	pt.y = rc.top + rc.bottom >> 1;
	convscale(in_reso, disp_reso, &pt);
	if((n = getnearframeno(frame, &pt)) >= 0){
		if(addmpstroke(&mp[n], op, pt_count)){
			free(op);
			return 1;
		}
	}else
		free(op);
	return 0;
}

MOJI_PATTERN *loadmparray(HOPDB hopdb,
		int page_No, FRAME *frame, OPDB_XY *in_reso, OPDB_XY *disp_reso){
	unsigned i, frame_count, st_count, err;
	HPAGEPAT hpp;
	MOJI_PATTERN *mp;
	frame_count = frame->count.x * frame->count.y;
	if((mp = makemparray(frame_count)) == NULL)
		return NULL;
	err = (hpp = LoadOPDBPagePattern(hopdb, page_No)) == NULL;
	st_count = GetOPDBStrokeCount(hpp);
	for(i = 0; i < st_count && !err; i++)
		err = addstroke(hpp, i, mp, frame, in_reso, disp_reso);
	if(hpp) ReleaseOPDBPagePattern(hpp);
	if(err){
		freemparray(mp, frame_count);
		mp = NULL;
	}
	return mp;
}


