/*
-----------------------------------------------------------------------------

					IPDB library for online pattern

											95/ 5/ 3	programmed by K.A

-----------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include "opdb.h"
#include "handle.h"

enum{
	XYByte = 1, XYWord = 2, XYDWord = 3,
	XSign = 0, XUnsign = 0x4, XMinus = 0xc,
	YSign = 0, YUnsign = 0x10, YMinus = 0x30
};

enum{
	HT_OPDB = 100, HT_Page
};

typedef struct tagOPDB_STROKE{
	void *add_info;
	WORD point_count;
	OPDB_XY *point;
	void *point_ai;
	struct tagOPDB_STROKE *lower, *upper;
} OPDB_STROKE;

struct tagPAGE_PAT_INFO;

typedef struct tagOPDB_INFO{
	HANDLE_LIST handle_info;
	HIPDB hipdb;
	int rflag;
	HAIFMT set_haif, page_haif, stroke_haif, point_haif;
	int set_aisize, page_aisize, stroke_aisize, point_aisize;
	struct tagPAGE_PAT_INFO *page_chain;
} OPDB_INFO;

typedef struct tagPAGE_PAT_INFO{
	HANDLE_LIST handle_info;
	OPDB_INFO *opdb_info;
	int del_flag;
	void *add_info;
	WORD page_No, stroke_count;
	OPDB_STROKE *stroke;
	struct tagPAGE_PAT_INFO *chain;
} PAGE_PAT_INFO;

static int _opdb_err_no = 0;
static const char _opdb_name[] = "online pattern data";
static const char _set_ai_name[] = "set additional information";
static const char _aif_name[] = "additional information format";

static OPDB_INFO *getopdbinfo(HOPDB hopdb){
	OPDB_INFO *oi;
	oi = handletopt(hopdb, HT_OPDB);
	if(oi != NULL && oi->rflag)
		oi = NULL;
	return oi;
}

static OPDB_INFO *getopdbinfoex(HOPDB hopdb){
	return (OPDB_INFO *)handletopt(hopdb, HT_OPDB);
}

static PAGE_PAT_INFO *getpageinfo(HPAGEPAT hpp){
	PAGE_PAT_INFO *ppi;
	ppi = handletopt(hpp, HT_Page);
	if(ppi != NULL && ppi->del_flag)
		ppi = NULL;
	return ppi;
}

static PAGE_PAT_INFO *getpageinfoex(HPAGEPAT hpp){
	return (PAGE_PAT_INFO *)handletopt(hpp, HT_Page);
}

/* p^[f[^x[X̃t@C𓾂 */
int GetOPDBFileName(HOPDB hopdb, char fn_buf[], int buf_size){
	OPDB_INFO *oi;
	if((oi = getopdbinfo(hopdb)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	return (_opdb_err_no = GetIPDBFileName(oi->hipdb, fn_buf, buf_size));
}

HAIFMT copyaif(HAIFMT haif){
	int i, n;
	HAIFMT new;
	char name[MaxAIItemNameLen];
	_opdb_err_no = 0;
	if((new = CreateAddInfoFormat()) == NULL){
		_opdb_err_no = GetAIFError();
		return NULL;
	}
	n = GetAddInfoItemCount(haif);
	for(i = 1; i <= n; i++){
		GetAddInfoItemName(haif, i, name, sizeof(name));
		if((_opdb_err_no = AddAddInfoItem(new, name,
				GetAddInfoItemType(haif, i),
				GetAddInfoItemArrayCount(haif, i))) != 0){
			ReleaseAddInfoFormat(new);
			return NULL;
		}
	}
	return new;
}

/* p^[f[^x[X̕ttH[}bg̃Rs[𓾂 */
HAIFMT GetOPDBAddInfoFormat(HOPDB hopdb, int kind){
	OPDB_INFO *oi;
	HAIFMT haif;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	switch(kind){
	case AIF_Set:
		haif = oi->set_haif;
		break;
	case AIF_Page:
		haif = oi->page_haif;
		break;
	case AIF_Stroke:
		haif = oi->stroke_haif;
		break;
	case AIF_Point:
		haif = oi->point_haif;
		break;
	default:
		_opdb_err_no = OPDB_IllegalValue;
		return NULL;
	}
	return copyaif(haif);
}

static OPDB_STROKE *getstroke(
		OPDB_STROKE *stroke, unsigned stroke_count, unsigned stroke_No){
	unsigned center = stroke_count >> 1;
	if(stroke_No >= stroke_count)
		return NULL;
	while(stroke != NULL){
		if(center > stroke_No){
			stroke = stroke->lower;
			stroke_count >>= 1;
			center -= ((stroke_count + 1) >> 1);
		}else if(center < stroke_No){
			stroke = stroke->upper;
			stroke_count = ((stroke_count + 1) >> 1) - 1;
			center += (stroke_count >> 1) + 1;
		}else
			break;
	}
	return stroke;
}

static void freestroke(OPDB_STROKE *stroke){
	if(stroke->point != NULL)
		free(stroke->point);
	free(stroke);
}

static void freestroketree(OPDB_STROKE *stroke_tree){
	if(stroke_tree == NULL)
		return;
	freestroketree(stroke_tree->lower);
	freestroketree(stroke_tree->upper);
	freestroke(stroke_tree);
}

static OPDB_STROKE *addstroke(OPDB_STROKE *, unsigned, OPDB_STROKE *, unsigned);
static OPDB_STROKE *removestroke(
		OPDB_STROKE *stroke_tree, unsigned stroke_count, unsigned stroke_No){
	unsigned center = stroke_count >> 1;
	if(stroke_tree == NULL || center == 0)
		return NULL;
	if(stroke_No < center){
		if(stroke_count & 1){
			OPDB_STROKE *tmp1, *tmp2;
			tmp1 = stroke_tree->upper;
			tmp2= removestroke(stroke_tree->lower, center, stroke_No);
			tmp2 = addstroke(tmp2, center - 1, stroke_tree, center - 1);
			stroke_tree = getstroke(tmp1, center, 0);
			stroke_tree->upper = removestroke(tmp1, center, 0);
			stroke_tree->lower = tmp2;
		}else
			stroke_tree->lower =
					removestroke(stroke_tree->lower, center, stroke_No);
	}else if(stroke_No > center){
		if(stroke_count & 1)
			stroke_tree->upper = removestroke(
					stroke_tree->upper, center, stroke_No - center - 1);
		else{
			OPDB_STROKE *tmp1, *tmp2;
			tmp1 = stroke_tree->lower;
			tmp2 = removestroke(
					stroke_tree->upper, center - 1, stroke_No - center - 1);
			tmp2 = addstroke(tmp2, center - 2, stroke_tree, 0);
			stroke_tree = getstroke(tmp1, center, center - 1);
			stroke_tree->lower = removestroke(tmp1, center, center - 1);
			stroke_tree->upper = tmp2;
		}
	}else{
		if(stroke_count & 1){
			OPDB_STROKE *tmp1, *tmp2;
			tmp2 = stroke_tree->lower;
			stroke_tree = getstroke(tmp1 = stroke_tree->upper, center, 0);
			stroke_tree->upper = removestroke(tmp1, center, 0);
			stroke_tree->lower = tmp2;
		}else{
			OPDB_STROKE *tmp1, *tmp2;
			tmp2 = stroke_tree->upper;
			stroke_tree =
					getstroke(tmp1 = stroke_tree->lower, center, center - 1);
			stroke_tree->lower = removestroke(tmp1, center, center - 1);
			stroke_tree->upper = tmp2;
		}
	}
	return stroke_tree;
}

static OPDB_STROKE *addstroke(OPDB_STROKE *stroke_tree,
		unsigned stroke_count, OPDB_STROKE *stroke, unsigned stroke_No){
	unsigned center = stroke_count >> 1;
	if(stroke_tree == NULL){
		stroke->lower = NULL;
		stroke->upper = NULL;
		return stroke;
	}
	if(stroke_No < center){
		if(stroke_count & 1)
			stroke_tree->lower =
					addstroke(stroke_tree->lower, center, stroke, stroke_No);
		else{
			OPDB_STROKE *tmp1, *tmp2;
			tmp1 = stroke_tree->lower;
			tmp2 = addstroke(stroke_tree->upper, center - 1, stroke_tree, 0);
			stroke_tree = getstroke(tmp1, center, center - 1);
			tmp1 = removestroke(tmp1, center, center - 1);
			stroke_tree->lower =
					addstroke(tmp1, center - 1, stroke, stroke_No);
			stroke_tree->upper = tmp2;
		}
	}else if(stroke_No > center){
		if(stroke_count & 1){
			OPDB_STROKE *tmp1, *tmp2;
			tmp1 = stroke_tree->upper;
			tmp2 = addstroke(
					stroke_tree->lower, center, stroke_tree, center);
			tmp1 = addstroke(tmp1, center, stroke, stroke_No - center - 1);
			stroke_tree = getstroke(tmp1, center + 1, 0);
			stroke_tree->upper = removestroke(tmp1, center + 1, 0);
			stroke_tree->lower = tmp2;
		}else
			stroke_tree->upper = addstroke(stroke_tree->upper,
					center - 1, stroke, stroke_No - center - 1);
	}else{
		if(stroke_count & 1){
			stroke_tree->lower = addstroke(
					stroke_tree->lower, center, stroke, stroke_No);
		}else{
			stroke->lower = stroke_tree->lower;
			stroke->upper = addstroke(
					stroke_tree->upper, center - 1, stroke_tree, 0);
			stroke_tree = stroke;
		}
	}
	return stroke_tree;
}

/* y[Wt𓾂 */
int GetOPDBPageAddInfo(HPAGEPAT hpp, int offset, void *data, int size){
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(offset < 0 || size < 0 ||
			offset + size > ppi->opdb_info->page_aisize)
		return (_opdb_err_no = OPDB_IllegalValue);
	memcpy(data, (BYTE *)ppi->add_info + offset, size);
	return 0;
}

/* y[Wtݒ肷 */
int SetOPDBPageAddInfo(HPAGEPAT hpp, int offset, void *data, int size){
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(offset < 0 || size < 0 ||
			offset + size > ppi->opdb_info->page_aisize)
		return (_opdb_err_no = OPDB_IllegalValue);
	memcpy((BYTE *)ppi->add_info + offset, data, size);
	return 0;
}

/* Xg[N̐𓾂 */
unsigned GetOPDBStrokeCount(HPAGEPAT hpp){
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return 0;
	}
	return ppi->stroke_count;
}

static OPDB_POINT *allocpoints(unsigned point_num, int aisize){
	unsigned i;
	BYTE *aip;
	OPDB_POINT *mp, *mptmp;
	if((mp = malloc((sizeof(OPDB_POINT) + aisize) * point_num)) == NULL)
		return NULL;
	if(aisize > 0){
		aip = (BYTE *)mp + sizeof(OPDB_POINT) * point_num;
		for(i = 0, mptmp = mp; i < point_num; i++, mptmp++, aip += aisize)
			mptmp->add_info = (void *)aip;
	}else
		for(i = 0, mptmp = mp; i < point_num; i++, mptmp++)
			mptmp->add_info = NULL;
	return mp;
}

static void transferpoints(OPDB_POINT *dst,
		OPDB_XY *src, void *point_ai, int aisize, unsigned point_count){
	unsigned i;
	for(i = 0; i < point_count; i++, src++, dst++){
		dst->x = src->x;
		dst->y = src->y;
		memcpy(dst->add_info, point_ai, aisize);
		(BYTE *)point_ai += aisize;
	}
}

static void transferpointsxy(OPDB_XY *dst, void *dst_ai,
		OPDB_XY *src, void *point_ai, int aisize, unsigned point_count){
	unsigned i;
	if(dst != NULL)
		for(i = 0; i < point_count; i++)
			*dst++ = *src++;
	if(dst_ai != NULL && point_ai != NULL)
		memcpy(dst_ai, point_ai, aisize * point_count);
}

static OPDB_XY *copypoints(OPDB_POINT *src, int aisize, unsigned point_num){
	long i;
	OPDB_XY *dst, *cur;
	void *ai;
	if((cur = dst = malloc((sizeof(*dst) + aisize) * point_num)) == NULL)
		return NULL;
	ai = dst + point_num;
	for(i = 0; i < (long)point_num; i++, cur++, src++){
		cur->x = src->x;
		cur->y = src->y;
		memcpy(ai, src->add_info, aisize);
		(BYTE *)ai += aisize;
	}
	return dst;
}

static OPDB_XY *copypointsxy(OPDB_XY *src,
		void *add_info, int aisize, unsigned point_num){
	long i;
	OPDB_XY *dst, *cur;
	if((cur = dst = malloc((sizeof(*dst) + aisize) * point_num)) == NULL)
		return NULL;
	for(i = 0; i < (long)point_num; i++)
		*cur++ = *src++;
	if(add_info != NULL)
		memcpy(cur, add_info, aisize * point_num);
	else
		memset(cur, 0, aisize * point_num);
	return dst;
}

/* |Cgz𓾂 */
int GetOPDBPoints(HPAGEPAT hpp, unsigned stroke_No,
		unsigned point_No, OPDB_POINT *pt_buf, unsigned pt_count){
	int aisize;
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(point_No >= ms->point_count || point_No + pt_count > ms->point_count)
		return (_opdb_err_no = OPDB_IllegalValue);
	aisize = ppi->opdb_info->point_aisize;
	transferpoints(pt_buf, ms->point + point_No,
			(BYTE *)ms->point_ai + aisize * point_No, aisize, pt_count);
	return 0;
}

/* Xg[N𓾂 */
int GetOPDBStrokeXY(HPAGEPAT hpp,
		unsigned stroke_No, OPDB_XY *xy, void *add_info, unsigned buf_count){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	transferpointsxy(xy, add_info, ms->point, ms->point_ai,
			ppi->opdb_info->point_aisize, min(buf_count, ms->point_count));
	return 0;
}

/* |Cgz𓾂 */
int GetOPDBPointsXY(HPAGEPAT hpp, unsigned stroke_No,
		unsigned point_No, OPDB_XY *xy, void *add_info, unsigned pt_count){
	int aisize;
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(point_No >= ms->point_count || point_No + pt_count > ms->point_count)
		return (_opdb_err_no = OPDB_IllegalValue);
	aisize = ppi->opdb_info->point_aisize;
	transferpointsxy(xy, add_info, ms->point + point_No,
			(BYTE *)ms->point_ai + aisize * point_No, aisize, pt_count);
	return 0;
}

/* Xg[N𓾂 */
int GetOPDBStroke(HPAGEPAT hpp,
		unsigned stroke_No, OPDB_POINT pt_buf[], unsigned buf_count){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	transferpoints(pt_buf, ms->point, ms->point_ai,
			ppi->opdb_info->point_aisize,
			min(ms->point_count, buf_count));
	return 0;
}

/* Xg[Nݒ肷 */
int SetOPDBStrokeXY(HPAGEPAT hpp,
		unsigned stroke_No, OPDB_XY *xy, void *add_info, unsigned point_count){
	OPDB_STROKE *ms;
	OPDB_XY *mpnew;
	PAGE_PAT_INFO *ppi;
	int aisize;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(!point_count || (ms =
			getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(xy == NULL)
		xy = ms->point;
	if(add_info == NULL)
		add_info = ms->point_ai;
	if((mpnew = copypointsxy(xy, add_info,
			aisize = ppi->opdb_info->point_aisize, point_count)) == NULL)
		return (_opdb_err_no = OPDB_MemoryAllocError);
	free(ms->point);
	ms->point_count = point_count;
	ms->point = mpnew;
	ms->point_ai = aisize ? mpnew + point_count: NULL;
	return 0;
}

/* Xg[Nݒ肷 */
int SetOPDBStroke(HPAGEPAT hpp,
		unsigned stroke_No, OPDB_POINT *mp, unsigned point_num){
	OPDB_STROKE *ms;
	OPDB_XY *mpnew;
	PAGE_PAT_INFO *ppi;
	int aisize;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(!point_num || (ms =
			getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	if((mpnew = copypoints(mp,
			aisize = ppi->opdb_info->point_aisize, point_num)) == NULL)
		return (_opdb_err_no = OPDB_MemoryAllocError);
	free(ms->point);
	ms->point_count = point_num;
	ms->point = mpnew;
	ms->point_ai = aisize ? mpnew + point_num: NULL;
	return 0;
}

/* Xg[N폜 */
int DeleteOPDBStroke(HPAGEPAT hpp, unsigned stroke_No){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	ppi->stroke = removestroke(ppi->stroke, ppi->stroke_count, stroke_No);
	freestroke(ms);
	ppi->stroke_count--;
	return 0;
}

static OPDB_STROKE *allocstroke(int ai_size){
	OPDB_STROKE *ms;
	if((ms = malloc(sizeof(OPDB_STROKE) + ai_size)) == NULL)
		return NULL;
	if(ai_size > 0)
		ms->add_info = (void *)((BYTE *)ms + sizeof(OPDB_STROKE));
	else
		ms->add_info = NULL;
	ms->point = ms->point_ai = NULL;
	return ms;
}

/* Xg[Nǉ */
int AddOPDBStrokeXY(HPAGEPAT hpp,
		unsigned stroke_No, OPDB_XY *xy, void *add_info, unsigned point_count){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	OPDB_XY *pt;
	int stroke_aisize, point_aisize;
	_opdb_err_no = 0;
	if(point_count <= 0)
		return (_opdb_err_no = OPDB_IllegalValue);
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	stroke_aisize = ppi->opdb_info->stroke_aisize;
	point_aisize = ppi->opdb_info->point_aisize;
	if((ms = allocstroke(stroke_aisize)) == NULL)
		return (_opdb_err_no = OPDB_MemoryAllocError);
	ms->point_count = point_count;
	if((pt = copypointsxy(xy,
			add_info, point_aisize, point_count)) == NULL){
		free(ms);
		return (_opdb_err_no = OPDB_MemoryAllocError);
	}
	ms->point = pt;
	ms->point_ai = point_aisize ? pt + point_count: NULL;
	ppi->stroke = addstroke(ppi->stroke, ppi->stroke_count, ms, stroke_No);
	ppi->stroke_count++;
	return 0;
}

/* Xg[Nǉ */
int AddOPDBStroke(HPAGEPAT hpp,
		unsigned stroke_No, OPDB_POINT *mp, unsigned point_num){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	OPDB_XY *xy;
	int stroke_aisize, point_aisize;
	_opdb_err_no = 0;
	if(point_num <= 0)
		return (_opdb_err_no = OPDB_IllegalValue);
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	stroke_aisize = ppi->opdb_info->stroke_aisize;
	point_aisize = ppi->opdb_info->point_aisize;
	if((ms = allocstroke(stroke_aisize)) == NULL)
		return (_opdb_err_no = OPDB_MemoryAllocError);
	ms->point_count = point_num;
	if((xy = copypoints(mp, point_aisize, point_num)) == NULL){
		free(ms);
		return (_opdb_err_no = OPDB_MemoryAllocError);
	}
	ms->point = xy;
	ms->point_ai = point_aisize ? xy + point_num: NULL;
	ppi->stroke = addstroke(ppi->stroke, ppi->stroke_count, ms, stroke_No);
	ppi->stroke_count++;
	return 0;
}

/* Xg[Nt𓾂 */
int GetOPDBStrokeAddInfo(HPAGEPAT hpp,
		unsigned stroke_No, int offset, void *data, int size){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(offset < 0 || size < 0 ||
			offset + size > ppi->opdb_info->stroke_aisize)
		return (_opdb_err_no = OPDB_IllegalValue);
	memcpy(data, (BYTE *)ms->add_info + offset, size);
	return 0;
}

/* Xg[Ntݒ肷 */
int SetOPDBStrokeAddInfo(HPAGEPAT hpp,
		unsigned stroke_No, int offset, void *data, int size){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(offset < 0 || size < 0 ||
			offset + size > ppi->opdb_info->stroke_aisize)
		return (_opdb_err_no = OPDB_IllegalValue);
	memcpy((BYTE *)ms->add_info + offset, data, size);
	return 0;
}

/* Xg[Ñ|Cg𓾂 */
unsigned GetOPDBPointCount(HPAGEPAT hpp, unsigned stroke_No){
	OPDB_STROKE *ms;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return 0;
	}
	if((ms = getstroke(ppi->stroke, ppi->stroke_count, stroke_No)) == NULL){
		_opdb_err_no = OPDB_IllegalValue;
		return 0;
	}
	return ms->point_count;
}

/* |Cg^z擾 */
OPDB_POINT *AllocOPDBPoints(HOPDB hopdb, unsigned point_num){
	int point_aisize;
	OPDB_INFO *oi;
	OPDB_POINT *mp;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	point_aisize = oi->point_aisize;
	if((mp = allocpoints(point_num, point_aisize)) == NULL){
		_opdb_err_no = OPDB_MemoryAllocError;
		return NULL;
	}
	return mp;
}

/* |Cg^zp */
void FreeOPDBPoints(OPDB_POINT *mp){
	_opdb_err_no = 0;
	if(mp != NULL)
		free(mp);
}

/* y[Wp^[ */
void ReleaseOPDBPagePattern(HPAGEPAT hpp){
	unsigned page_No;
	PAGE_PAT_INFO *ppi, **tmp;
	_opdb_err_no = 0;
	if((ppi = getpageinfoex(hpp)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return;
	}
	page_No = ppi->page_No;
	freestroketree(ppi->stroke);
	tmp = &ppi->opdb_info->page_chain;
	while((ppi = *tmp) != NULL){
		if(ppi->page_No == page_No)
			break;
		tmp = &ppi->chain;
	}
	if(ppi != NULL){
		*tmp = ppi->chain;
		if(ppi->opdb_info->rflag)
			ReleaseOPDB(ppi->opdb_info->handle_info.handle);
	}
	if(ppi->add_info)
		free(ppi->add_info);
	removehandleinfo(hpp, HT_Page);
}

/* y[Wp^[ɑ΂p^[f[^x[X̃nh𓾂 */
HOPDB GetPagePatternOPDB(HPAGEPAT hpp){
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	return ppi->opdb_info->handle_info.handle;
}

/* y[Wp^[̃y[Wԍ𓾂 */
unsigned GetOPDBPageNo(HPAGEPAT hpp){
	PAGE_PAT_INFO *ppi;
	if((ppi = getpageinfo(hpp)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return 0;
	}
	return ppi->page_No;
}

static PAGE_PAT_INFO *getalreadyppi(PAGE_PAT_INFO *ppi, int page){
	while(ppi != NULL){
		if(ppi->page_No == page && !ppi->del_flag)
			break;
		ppi = ppi->chain;
	}
	return ppi;
}

static void clearai(void *data, int size){
	int i;
	BYTE *pt = (BYTE *)data;
	for(i = 0; i < size; i++, pt++)
		*pt = 0;
}

static PAGE_PAT_INFO *createppi(OPDB_INFO *oi, unsigned page){
	PAGE_PAT_INFO *ppi;
	void *tmp = NULL;
	int page_aisize = oi->page_aisize;

	_opdb_err_no = 0;
	if(page >= GetIPDBIndexPageCount(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _opdb_name))){
		_opdb_err_no = OPDB_IllegalValue;
		return NULL;
	}
	if(getalreadyppi(oi->page_chain, page) != NULL){
		_opdb_err_no = OPDB_AlreadyExist;
		return NULL;
	}
	if(page_aisize > 0){
		if((tmp = malloc(page_aisize)) == NULL){
			_opdb_err_no = OPDB_MemoryAllocError;
			return NULL;
		}
		clearai(tmp, page_aisize);
	}
	if((ppi = makehandleinfo(HT_Page, sizeof(*ppi))) == NULL){
		if(tmp)
			free(tmp);
		_opdb_err_no = (gethandlecount() ==
				MaxHandleCount) ? IPDB_HandleOver: IPDB_MemoryAllocError;
		return NULL;
	}
	ppi->add_info = tmp;
	ppi->opdb_info = oi;
	ppi->chain = oi->page_chain;
	oi->page_chain = ppi;
	ppi->page_No = page;
	ppi->stroke_count = 0;
	ppi->stroke = NULL;
	ppi->del_flag = 0;
	return ppi;
}

/* y[Wp^[VK쐬 */
HPAGEPAT CreateOPDBPagePattern(HOPDB hopdb, unsigned page){
	OPDB_INFO *oi;
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	if((ppi = createppi(oi, page)) == NULL)
		return NULL;
	return ppi->handle_info.handle;
}

static int readstroke(HIPDB hipdb,
		OPDB_STROKE *ms, int stroke_aisize, int point_aisize){
	unsigned i;
	BYTE type;
	OPDB_XY *mp, *cur, pt;
	void *ai;
	_opdb_err_no = 0;
	if(stroke_aisize > 0)
		ReadIPDBPage(hipdb, ms->add_info, stroke_aisize);
	if(ReadIPDBPage(hipdb, &ms->point_count, sizeof(ms->point_count)))
		return (_opdb_err_no = GetIPDBError());
	if(ms->point_count <= 0)
		return (_opdb_err_no = OPDB_IllegalFile);
	ReadIPDBPage(hipdb, &pt, sizeof(pt));
	if((cur = mp = malloc((sizeof(*mp) +
			point_aisize) * ms->point_count)) == NULL)
		return (_opdb_err_no = OPDB_MemoryAllocError);
	ai = mp + ms->point_count;
	ms->point_ai = (point_aisize > 0) ? mp + ms->point_count: NULL;
	if(point_aisize > 0)
		ReadIPDBPage(hipdb, ai, point_aisize);
	cur->x = pt.x;
	cur->y = pt.y;
	cur++;
	(BYTE *)ai += point_aisize;
	if(ms->point_count <= 1){
		ms->point = mp;
		return 0;
	}
	if(ReadIPDBPage(hipdb, &type, sizeof(type))){
		free(mp);
		return (_opdb_err_no = GetIPDBError());
	}
	if(type == XYByte){
		unsigned char xy;
		int d;
		for(i = 1; i < ms->point_count; i++, cur++){
			if(ReadIPDBPage(hipdb, &xy, sizeof(xy)) ||
					(point_aisize > 0 &&
					ReadIPDBPage(hipdb, ai, point_aisize))){
				free(mp);
				return (_opdb_err_no = GetIPDBError());
			}
			pt.x += ((d = xy >> 4) & 8) ? d - 0x10: d;
			pt.y += ((d = xy & 0xf) & 8) ? d - 0x10: d;
			cur->x = pt.x; cur->y = pt.y;
			(BYTE *)ai += point_aisize;
		}
	}else if(type == XYWord){
		struct {
			char x, y;
		} xy;
		for(i = 1; i < ms->point_count; i++, cur++){
			if(ReadIPDBPage(hipdb, &xy, sizeof(xy)) ||
					(point_aisize > 0 &&
					ReadIPDBPage(hipdb, ai, point_aisize))){
				free(mp);
				return (_opdb_err_no = GetIPDBError());
			}
			pt.x += xy.x; pt.y += xy.y;
			cur->x = pt.x; cur->y = pt.y;
			(BYTE *)ai += point_aisize;
		}
	}else if(type == XYDWord){
		OPDB_XY xy;
		for(i = 1; i < ms->point_count; i++, cur++){
			if(ReadIPDBPage(hipdb, &xy, sizeof(xy)) ||
					(point_aisize > 0 &&
					ReadIPDBPage(hipdb, ai, point_aisize))){
				free(mp);
				return (_opdb_err_no = OPDB_ReadError);
			}
			pt.x += xy.x; pt.y += xy.y;
			cur->x = pt.x; cur->y = pt.y;
			(BYTE *)ai += point_aisize;
		}
	}else{
		free(mp);
		return (_opdb_err_no = OPDB_IllegalFile);
	}
	ms->point = mp;
	return 0;
}

static int readpage(HIPDB hipdb, PAGE_PAT_INFO *ppi){
	unsigned i;
	int page_aisize, stroke_aisize, point_aisize;
	OPDB_STROKE *ms;

	_opdb_err_no = 0;
	page_aisize = ppi->opdb_info->page_aisize;
	stroke_aisize = ppi->opdb_info->stroke_aisize;
	point_aisize = ppi->opdb_info->point_aisize;
	if(page_aisize > 0)
		if(ReadIPDBPage(hipdb, ppi->add_info, page_aisize))
			return (_opdb_err_no = GetIPDBError());
	if(ReadIPDBPage(hipdb, &ppi->stroke_count, sizeof(ppi->stroke_count)))
		return (_opdb_err_no = GetIPDBError());
	for(i = 0; i < ppi->stroke_count; i++){
		if((ms = allocstroke(stroke_aisize)) == NULL){
			_opdb_err_no = OPDB_MemoryAllocError;
			break;
		}
		if(readstroke(hipdb, ms, stroke_aisize, point_aisize)){
			freestroke(ms);
			break;
		}
		ppi->stroke = addstroke(ppi->stroke, i, ms, i);
	}
	if(_opdb_err_no)
		freestroketree(ppi->stroke);
	return _opdb_err_no;
}

/* y[Wp^[ǂݍ */
HPAGEPAT LoadOPDBPagePattern(HOPDB hopdb, unsigned page){
	OPDB_INFO *oi;
	PAGE_PAT_INFO *ppi;

	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	if((ppi = createppi(oi, page)) == NULL)
		return NULL;
	if(!ReadOpenIPDBPage(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _opdb_name), page)){
		if(GetIPDBPageSize(oi->hipdb) > 0)
			readpage(oi->hipdb, ppi);
		CloseIPDBPage(oi->hipdb);
	}else
		_opdb_err_no = GetIPDBError();
	if(_opdb_err_no){
		int tmp = _opdb_err_no;
		ReleaseOPDBPagePattern(ppi->handle_info.handle);
		_opdb_err_no = tmp;
		return NULL;
	}
	return ppi->handle_info.handle;
}

static unsigned char getpointtype(OPDB_XY *mp, unsigned point_count){
	unsigned i;
	int dx, dy;
	OPDB_XY prev;
	unsigned char type;

	prev.x = mp->x; prev.y = mp->y;
	mp++;
	type = XYByte;
	for(i = 1; i < point_count; i++){
		dx = mp->x - prev.x; dy = mp->y - prev.y;
		if(dx > 127 || dx < -128 || dy > 127 || dy < -128){
			type = XYDWord;
			break;
		}else if(dx > 7 || dx < -8 || dy > 7 || dy < -8)
			type = XYWord;
		prev.x = mp->x; prev.y = mp->y;
		mp++;
	}
	return type;
}

static long getpagesize(PAGE_PAT_INFO *ppi){
	unsigned i;
	int page_aisize, stroke_aisize, point_aisize;
	int type;
	long size;
	OPDB_INFO *oi;
	OPDB_STROKE *st;

	oi = ppi->opdb_info;
	page_aisize = oi->page_aisize;
	stroke_aisize = oi->stroke_aisize;
	point_aisize = oi->point_aisize;
	size = page_aisize + sizeof(ppi->stroke_count);
	for(i = 0; i < ppi->stroke_count; i++){
		st = getstroke(ppi->stroke, ppi->stroke_count, i);
		size += (long)st->point_count * point_aisize +
				sizeof(st->point_count) + sizeof(SHORT) * 2 + stroke_aisize;
		if(st->point_count <= 1)
			continue;
		type = getpointtype(st->point, st->point_count);
		size += (long)((type == XYByte) ?
				1: ((type == XYWord) ? 2: 4)) * (st->point_count - 1) +
				sizeof(CHAR);
	}
	return size;
}

static int writestroke(HIPDB hipdb,
		OPDB_STROKE *ms, int stroke_aisize, int point_aisize){
	unsigned i;
	BYTE type;
	OPDB_XY *mp, prev;
	void *ai;

	_opdb_err_no = 0;
	if(stroke_aisize > 0)
		if(WriteIPDBPage(hipdb, ms->add_info, stroke_aisize))
			return (_opdb_err_no = GetIPDBError());
	if(WriteIPDBPage(hipdb, &ms->point_count, sizeof(ms->point_count)))
		return (_opdb_err_no = GetIPDBError());
	mp = ms->point;
	ai = ms->point_ai;
	prev.x = mp->x;
	prev.y = mp->y;
	if(WriteIPDBPage(hipdb, &prev, sizeof(prev)))
		return (_opdb_err_no = GetIPDBError());
	if(point_aisize > 0){
		if(WriteIPDBPage(hipdb, ai, point_aisize))
			return (_opdb_err_no = GetIPDBError());
		(BYTE *)ai += point_aisize;
	}
	if(ms->point_count <= 1)
		return 0;
	type = getpointtype(mp, ms->point_count);
	if(WriteIPDBPage(hipdb, &type, sizeof(type)))
		return (_opdb_err_no = GetIPDBError());
	mp++;
	if(type == XYByte){
		BYTE xy;
		for(i = 1; i < ms->point_count; i++, mp++){
			xy = (((mp->x - prev.x) &
					0xf) << 4) | ((mp->y - prev.y) & 0xf);
			if(WriteIPDBPage(hipdb, &xy, sizeof(xy)) ||
					(point_aisize > 0 &&
					WriteIPDBPage(hipdb, ai, point_aisize)))
				return (_opdb_err_no = GetIPDBError());
			prev.x = mp->x; prev.y = mp->y;
			(BYTE *)ai += point_aisize;
		}
	}else if(type == XYWord){
		struct {
			CHAR x, y;
		} xy;
		for(i = 1; i < ms->point_count; i++, mp++){
			xy.x = mp->x - prev.x; xy.y = mp->y - prev.y;
			if(WriteIPDBPage(hipdb, &xy, sizeof(xy)) ||
					(point_aisize > 0 &&
					WriteIPDBPage(hipdb, ai, point_aisize)))
				return (_opdb_err_no = GetIPDBError());
			prev.x = mp->x; prev.y = mp->y;
			(BYTE *)ai += point_aisize;
		}
	}else if(type == XYDWord){
		OPDB_XY xy;
		for(i = 1; i < ms->point_count; i++, mp++){
			xy.x = mp->x - prev.x; xy.y = mp->y - prev.y;
			if(WriteIPDBPage(hipdb, &xy, sizeof(xy)) ||
					(point_aisize > 0 &&
					WriteIPDBPage(hipdb, ai, point_aisize)))
				return (_opdb_err_no = GetIPDBError());
			prev.x = mp->x; prev.y = mp->y;
			(BYTE *)ai += point_aisize;
		}
	}
	return 0;
}

static int writepage(HIPDB hipdb, PAGE_PAT_INFO *ppi){
	unsigned i;
	int page_aisize, stroke_aisize, point_aisize;

	_opdb_err_no = 0;
	page_aisize = ppi->opdb_info->page_aisize;
	stroke_aisize = ppi->opdb_info->stroke_aisize;
	point_aisize = ppi->opdb_info->point_aisize;
	if(page_aisize > 0)
		if(WriteIPDBPage(hipdb, ppi->add_info, page_aisize))
			return (_opdb_err_no = GetIPDBError());
	if(WriteIPDBPage(hipdb, &ppi->stroke_count, sizeof(ppi->stroke_count)))
		return (_opdb_err_no = GetIPDBError());
	for(i = 0; i < ppi->stroke_count; i++){
		if(writestroke(hipdb,
				getstroke(ppi->stroke, ppi->stroke_count, i),
				stroke_aisize, point_aisize))
			break;
	}
	return _opdb_err_no;
}

/* y[Wp^[ */
int SaveOPDBPagePattern(HPAGEPAT hpp){
	PAGE_PAT_INFO *ppi;
	_opdb_err_no = 0;
	if((ppi = getpageinfo(hpp)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(WriteOpenIPDBPage(ppi->opdb_info->hipdb,
			GetIPDBIndexNo(ppi->opdb_info->hipdb, _opdb_name),
			ppi->page_No, getpagesize(ppi)))
		return (_opdb_err_no = GetIPDBError());
	writepage(ppi->opdb_info->hipdb, ppi);
	CloseIPDBPage(ppi->opdb_info->hipdb);
	return _opdb_err_no;
}

static void freeopdbinfo(OPDB_INFO *oi){
	ReleaseAddInfoFormat(oi->set_haif);
	ReleaseAddInfoFormat(oi->page_haif);
	ReleaseAddInfoFormat(oi->stroke_haif);
	ReleaseAddInfoFormat(oi->point_haif);
	ReleaseIPDB(oi->hipdb);
}

/* p^[f[^x[X */
void ReleaseOPDB(HOPDB hopdb){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = getopdbinfoex(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return;
	}
	if(oi->page_chain != NULL){
		oi->rflag = 1;
		return;
	}
	freeopdbinfo(oi);
	removehandleinfo(hopdb, HT_OPDB);
}

static long getaifdatalen(HAIFMT haif){
	long size = 1;
	int i, n;
	char name[MaxAIItemNameLen];
	n = GetAddInfoItemCount(haif);
	for(i = 1; i <= n; i++){
		GetAddInfoItemName(haif, i, name, sizeof(name));
		size += (strlen(name) + 1) + 3;
	}
	return size;
}

static int writeaif(HIPDB hipdb, HAIFMT haif){
	int i, n;
	WORD ac;
	BYTE len, type;
	char name[MaxAIItemNameLen];
	_opdb_err_no = 0;
	n = GetAddInfoItemCount(haif);
	for(i = 1; i <= n; i++){
		GetAddInfoItemName(haif, i, name, sizeof(name));
		type = GetAddInfoItemType(haif, i);
		ac = GetAddInfoItemArrayCount(haif, i);
		len = strlen(name);
		if(WriteIPDBPage(hipdb, &len, sizeof(len)) ||
				WriteIPDBPage(hipdb, name, len) ||
				WriteIPDBPage(hipdb, &type, sizeof(type)) ||
				WriteIPDBPage(hipdb, &ac, sizeof(ac)))
			return (_opdb_err_no = GetIPDBError());
	}
	len = 0;
	if(WriteIPDBPage(hipdb, &len, sizeof(len)))
		return (_opdb_err_no = GetIPDBError());
	return 0;
}

static HAIFMT readaif(HIPDB hipdb){
	BYTE len, type;
	SHORT array_count;
	HAIFMT haif;
	char buf[MaxAIItemNameLen];

	if((haif = CreateAddInfoFormat()) == NULL){
		_opdb_err_no = GetAIFError();
		return NULL;
	}
	for(;;){
		if((_opdb_err_no =
				ReadIPDBPage(hipdb, &len, sizeof(len))) != 0 || !len)
			break;
		if(len > sizeof(buf) - 1){
			_opdb_err_no = OPDB_IllegalFile;
			break;
		}
		if(ReadIPDBPage(hipdb, buf, len) ||
				ReadIPDBPage(hipdb, &type, sizeof(type)) ||
				ReadIPDBPage(hipdb, &array_count, sizeof(array_count))){
			_opdb_err_no = GetIPDBError();
			break;
		}
		buf[len] = 0;
		if((_opdb_err_no = AddAddInfoItem(haif, buf, type, array_count)) != 0)
			break;
	}
	if(_opdb_err_no){
		ReleaseAddInfoFormat(haif);
		haif = NULL;
	}
	return haif;
}

static int createopdbfile(OPDB_INFO *oi, const char fn[], unsigned page_count){
	char tmp;
	int i;
	long size;
	_opdb_err_no = 0;
	if((oi->hipdb = CreateIPDB(fn)) == NULL)
		return (_opdb_err_no = GetIPDBError());
	if(MakeIPDBIndex(oi->hipdb, _aif_name, 1) ||
			MakeIPDBIndex(oi->hipdb, _set_ai_name, 1) ||
			MakeIPDBIndex(oi->hipdb, _opdb_name, page_count)){
		_opdb_err_no = GetIPDBError();
		goto err1;
	}
	size = getaifdatalen(oi->set_haif) + getaifdatalen(oi->page_haif) +
			getaifdatalen(oi->stroke_haif) + getaifdatalen(oi->point_haif);
	if(WriteOpenIPDBPage(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _aif_name), 0, size)){
		_opdb_err_no = GetIPDBError();
		goto err1;
	}
	if(writeaif(oi->hipdb, oi->set_haif) ||
			writeaif(oi->hipdb, oi->page_haif) ||
			writeaif(oi->hipdb, oi->stroke_haif) ||
			writeaif(oi->hipdb, oi->point_haif))
		goto err1;
	CloseIPDBPage(oi->hipdb);
	if(WriteOpenIPDBPage(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _set_ai_name), 0, oi->set_aisize)){
		_opdb_err_no = GetIPDBError();
		goto err1;
	}
	tmp = 0;
	for(i = 0; i < oi->set_aisize; i++)
		if(WriteIPDBPage(oi->hipdb, &tmp, 1)){
			_opdb_err_no = GetIPDBError();
			goto err1;
		}
	CloseIPDBPage(oi->hipdb);
	err1:
	if(_opdb_err_no){
		ReleaseIPDB(oi->hipdb);
		oi->hipdb = NULL;
		remove(fn);
	}
	return _opdb_err_no;
}

static OPDB_INFO *createopdb(void){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = makehandleinfo(HT_OPDB, sizeof(OPDB_INFO))) == NULL){
		_opdb_err_no = (gethandlecount() ==
				MaxHandleCount) ? IPDB_HandleOver: IPDB_MemoryAllocError;
		return NULL;
	}
	oi->rflag = 0;
	oi->hipdb = NULL;
	oi->set_haif = oi->page_haif = oi->stroke_haif = oi->point_haif = NULL;
	oi->set_aisize =
			oi->page_aisize = oi->stroke_aisize = oi->point_aisize = 0;
	oi->page_chain = NULL;
	return oi;
}

static HOPDB GetIPDBsOPDB(HIPDB hipdb){
	int aif_no;
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((aif_no = GetIPDBIndexNo(hipdb, _aif_name)) == NULL ||
			GetIPDBIndexNo(hipdb, _set_ai_name) == NULL ||
			GetIPDBIndexNo(hipdb, _opdb_name) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	if((oi = createopdb()) == NULL)
		return NULL;
	if(!ReadOpenIPDBPage(hipdb, aif_no, 0)){
		if((oi->set_haif = readaif(hipdb)) != NULL &&
				(oi->page_haif = readaif(hipdb)) != NULL &&
				(oi->stroke_haif = readaif(hipdb)) != NULL &&
				(oi->point_haif = readaif(hipdb)) != NULL){
			oi->set_aisize = GetAddInfoSize(oi->set_haif);
			oi->page_aisize = GetAddInfoSize(oi->page_haif);
			oi->stroke_aisize = GetAddInfoSize(oi->stroke_haif);
			oi->point_aisize = GetAddInfoSize(oi->point_haif);
		}
		CloseIPDBPage(hipdb);
	}else
		_opdb_err_no = GetIPDBError();
	if(_opdb_err_no){
		int tmp = _opdb_err_no;
		ReleaseOPDB(oi->handle_info.handle);
		_opdb_err_no = tmp;
		return NULL;
	}
	oi->hipdb = hipdb;
	return oi->handle_info.handle;
}

HIPDB GetOPDBsIPDB(HOPDB hopdb){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return NULL;
	}
	return oi->hipdb;
}

/* p^[f[^x[XVK쐬 */
HOPDB CreateOPDB(const char data_file[], HAIFMT set_haif, HAIFMT page_haif,
		HAIFMT stroke_haif, HAIFMT point_haif, unsigned page_count){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = createopdb()) == NULL)
		return NULL;
	if((oi->set_haif = copyaif(set_haif)) != NULL &&
			(oi->page_haif = copyaif(page_haif)) != NULL &&
			(oi->stroke_haif = copyaif(stroke_haif)) != NULL &&
			(oi->point_haif = copyaif(point_haif)) != NULL){
		oi->set_aisize = GetAddInfoSize(oi->set_haif);
		oi->page_aisize = GetAddInfoSize(oi->page_haif);
		oi->stroke_aisize = GetAddInfoSize(oi->stroke_haif);
		oi->point_aisize = GetAddInfoSize(oi->point_haif);
		createopdbfile(oi, data_file, page_count);
	}
	if(_opdb_err_no){
		int tmp = _opdb_err_no;
		ReleaseOPDB(oi->handle_info.handle);
		_opdb_err_no = tmp;
		return NULL;
	}
	return oi->handle_info.handle;
}

/* p^[f[^x[Xǂݍ */
HOPDB LoadOPDB(const char data_file[]){
	HIPDB hipdb;
	_opdb_err_no = 0;
	if((hipdb = LoadIPDB(data_file)) == NULL){
		_opdb_err_no = GetIPDBError();
		return NULL;
	}
	return GetIPDBsOPDB(hipdb);
}

/* y[W𓾂 */
unsigned GetOPDBPageCount(HOPDB hopdb){
	OPDB_INFO *oi;
	unsigned page_count;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return 0;
	}
	page_count = GetIPDBIndexPageCount(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _opdb_name));
	_opdb_err_no = GetIPDBError();
	return page_count;
}

static void addpagenumber(
		PAGE_PAT_INFO *ppi, unsigned page_pos, long page_count){
	while(ppi != NULL){
		if(ppi->page_No >= page_pos){
			if((long)ppi->page_No + page_count < (long)page_pos)
				ppi->del_flag = 1;
			else
				ppi->page_No += (WORD)page_count;
		}
		ppi = ppi->chain;
	}
}

/* y[W𑝂₷ */
int AddOPDBPages(HOPDB hopdb, unsigned page_pos, unsigned page_count){
	OPDB_INFO *oi;
	if((oi = getopdbinfo(hopdb)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(AddIPDBIndexPages(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _opdb_name), page_pos, page_count))
		return (_opdb_err_no = GetIPDBError());
	addpagenumber(oi->page_chain, page_pos, page_count);
	return (_opdb_err_no = 0);
}

/* y[W炷 */
int DeleteOPDBPages(HOPDB hopdb, unsigned page_pos, unsigned page_count){
	OPDB_INFO *oi;
	if((oi = getopdbinfo(hopdb)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(DeleteIPDBIndexPages(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _opdb_name), page_pos, page_count))
		return (_opdb_err_no = GetIPDBError());
	addpagenumber(oi->page_chain, page_pos, -(long)page_count);
	return (_opdb_err_no = 0);
}

/* p^[f[^x[X̕t𓾂 */
int LoadOPDBAddInfo(HOPDB hopdb, int offset, void *data, int size){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(offset < 0 || size < 0 ||
			offset >= oi->set_aisize || offset + size > oi->set_aisize)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(ReadOpenIPDBPage(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _set_ai_name), 0) ||
			SeekIPDBPage(oi->hipdb, offset) ||
			ReadIPDBPage(oi->hipdb, data, size))
		_opdb_err_no = GetIPDBError();
	CloseIPDBPage(oi->hipdb);
	return _opdb_err_no;
}

/* p^[f[^x[X̕tݒ肷 */
int SaveOPDBAddInfo(HOPDB hopdb, int offset, void *data, int size){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	if(offset < 0 || size < 0 ||
			offset >= oi->set_aisize || offset + size > oi->set_aisize)
		return (_opdb_err_no = OPDB_IllegalValue);
	if(UpdateOpenIPDBPage(oi->hipdb,
			GetIPDBIndexNo(oi->hipdb, _set_ai_name), 0) ||
			SeekIPDBPage(oi->hipdb, offset) ||
			WriteIPDBPage(oi->hipdb, data, size))
		_opdb_err_no = GetIPDBError();
	CloseIPDBPage(oi->hipdb);
	return _opdb_err_no;
}

/* f[^x[Xt@C̖gp̈̃TCY𓾂 */
long GetOPDBFlagSize(HOPDB hopdb){
	long size;
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL){
		_opdb_err_no = OPDB_IllegalHandle;
		return 0;
	}
	size = GetIPDBFlagSize(oi->hipdb);
	if((_opdb_err_no = GetIPDBError()) != 0)
		return 0;
	return size;
}

/* f[^x[Xt@C̖gp̈폜 */
int ArrangeOPDB(HOPDB hopdb){
	OPDB_INFO *oi;
	_opdb_err_no = 0;
	if((oi = getopdbinfo(hopdb)) == NULL)
		return (_opdb_err_no = OPDB_IllegalHandle);
	return (_opdb_err_no = ArrangeIPDB(oi->hipdb));
}

/* ڂG[bZ[W𓾂 */
int GetOPDBError(void){
	return _opdb_err_no;
}
