#include <stdio.h>
#include <stdlib.h>
//#include <alloc.h>
#include "btype.h"
#include "rlb.h"

static const char _rlb_name[] = "run length bitmap";

void *AllocRLB(void){
	return malloc((long)(RLBSizeX / 8) * RLBSizeY);
}

void FreeRLB(void *rlb){
	if(rlb != NULL)
		free(rlb);
}

static long getrlbsize(void *rlb){
	long i, j, size;
	BYTE pat;
	CHAR run = 0;
	for(i = j = size = 0; i < (long)(RLBSizeX / 8) * RLBSizeY; i++){
		if(j++ == RLBSizeX / 8)
			j = 1;
		pat = *((CHAR *)rlb)++;
		if(!pat){
			if(run < 0){
				size += 1 - run;
				run = 0;
			}
			run++;
			if(j == RLBSizeX / 8){
				size++;
				run = 0;
			}
		}else{
			if(run > 0){
				size++;
				run = 0;
			}
			run--;
			if(j == RLBSizeX / 8){
				size += 1 - run;
				run = 0;
			}
		}
	}
	return size;
}

int GetRLBIndexNo(HIPDB hipdb){
	return GetIPDBIndexNo(hipdb, _rlb_name);
}

int SaveRLB(HIPDB hipdb, int page_No, void *rlb){
	int err = 0;
	long i, j;
	CHAR run;
	BYTE buf[RLBSizeX / 8], pat;
	if(WriteOpenIPDBPage(hipdb,
			GetRLBIndexNo(hipdb), page_No, getrlbsize(rlb)))
		return GetIPDBError();
	run = 0;
	for(i = j = 0; i < (long)(RLBSizeX / 8) * RLBSizeY; i++){
		if(j++ == RLBSizeX / 8)
			j = 1;
		pat = *((BYTE *)rlb)++;
		if(!pat){
			if(run < 0){
				if(WriteIPDBPage(hipdb, &run, sizeof(char)) ||
						WriteIPDBPage(hipdb, buf, -run * sizeof(char))){
					err = GetIPDBError();
					break;
				}
				run = 0;
			}
			run++;
			if(j == RLBSizeX / 8){
				if(WriteIPDBPage(hipdb, &run, sizeof(char))){
					err = GetIPDBError();
					break;
				}
				run = 0;
			}
		}else{
			if(run > 0){
				if(WriteIPDBPage(hipdb, &run, sizeof(char))){
					err = GetIPDBError();
					break;
				}
				run = 0;
			}
			buf[-(run--)] = pat;
			if(j == RLBSizeX / 8){
				if(WriteIPDBPage(hipdb, &run, sizeof(char)) ||
						WriteIPDBPage(hipdb, buf, -run * sizeof(char))){
					err = GetIPDBError();
					break;
				}
				run = 0;
			}
		}
	}
	CloseIPDBPage(hipdb);
	return err;
}

int LoadRLB(HIPDB hipdb, int page_No, void *rlb){
	int err = 0;
	long i = 0, j;
	BYTE buf[RLBSizeX / 8];
	CHAR pat;
	if(ReadOpenIPDBPage(hipdb, GetRLBIndexNo(hipdb), page_No))
		return GetIPDBError();
	for(;;){
		if(ReadIPDBPage(hipdb, &pat, sizeof(pat)) ||
				i >= (long)(RLBSizeX / 8) * RLBSizeY){
			err = GetIPDBError();
			break;
		}
		if(pat > 0){
			if(i + pat > (long)(RLBSizeX / 8) * RLBSizeY){
				err = IPDB_IllegalFile;
				break;
			}
			for(j = 0; j < pat; j++)
				*((CHAR *)rlb + i++) = 0;
		}else if(pat < 0){
			if(-pat > sizeof(buf)){
				err = IPDB_IllegalFile;
                break;
			}
			if(ReadIPDBPage(hipdb, buf, -pat)){
				err = GetIPDBError();
				break;
			}
			if(i - pat > (long)(RLBSizeX / 8) * RLBSizeY){
				err = IPDB_IllegalFile;
				break;
			}
			for(j = 0; j < -pat; j++)
				*((CHAR *)rlb + i++) = buf[j];
		}else
			break;
	}
	CloseIPDBPage(hipdb);
	if(i != (long)(RLBSizeX / 8) * RLBSizeY)
		err = IPDB_IllegalFile;
	return err;
}

int MakeRLBIndex(HIPDB hipdb, int page_count){
	return MakeIPDBIndex(hipdb, _rlb_name, page_count);
}

int RemoveRLBIndex(HIPDB hipdb){
	return RemoveIPDBIndex(hipdb, _rlb_name);
}

int AddRLBPages(HIPDB hipdb, int page_No, int page_count){
	if(page_No < 0 || page_count < 0)
		return IPDB_IllegalValue;
	return AddIPDBIndexPages(hipdb, GetRLBIndexNo(hipdb), page_No, page_count);
}

int DeleteRLBPages(HIPDB hipdb, int page_No, int page_count){
	if(page_No < 0 || page_count < 0)
		return IPDB_IllegalValue;
	return DeleteIPDBIndexPages(
			hipdb, GetRLBIndexNo(hipdb), page_No, page_count);
}
