#include <stdio.h>
#include <string.h>
#include <io.h>
#include <time.h>
#include "bctovc.h"
#include "dir.h"

char *_tmpnam(char *buf){
	static char str[9];
	if(buf == NULL)
		buf = str;
	strcpy(buf, ")_(XXXXX");
	_mktemp(buf);
	return buf;
}

void fnmerge(char *buf,
		const char *drive, const char *dir, const char *fn, const char *ext){
	buf[0] = 0;
	if(drive){
		strcat(buf, drive);
		if(drive[0] && drive[strlen(drive) - 1] != ':')
			strcat(buf, ":");
	}
	if(dir){
		strcat(buf, dir);
		if(dir[0] && dir[strlen(dir) - 1] != '\\')
			strcat(buf, "\\");
	}
	if(fn)
		strcat(buf, fn);
	if(ext){
		if(ext[0] != 0 && ext[0] != '.')
			strcat(buf, ".");
		strcat(buf, ext);
	}
}

void fnsplit(const char *buf,
		char *drive, char *dir, char *fn, char *ext){
	int l;
	int dip, fnp, exp;
	l = strlen(buf);
	for(fnp = l - 1; fnp >= 0 &&
			buf[fnp] != ':' && buf[fnp] != '\\'; fnp--);
	fnp++;
	for(exp = fnp; buf[exp] != 0 && buf[exp] !='.'; exp++);
	for(dip = 0; buf[dip] != '\\' &&
			buf[dip] != 0 && buf[dip] != ':'; dip++);
	if(buf[dip] != ':')
		dip = -1;
	dip++;
	if(drive){
		memcpy(drive, buf, dip);
		drive[dip] = 0;
	}
	if(dir){
		memcpy(dir, &buf[dip], fnp - dip);
		dir[fnp - dip] = 0;
	}
	if(fn){
		memcpy(fn, &buf[fnp], exp - fnp);
		fn[exp - fnp] = 0;
	}
	if(ext){
		memcpy(ext, &buf[exp], l - exp);
		ext[l - exp] = 0;
	}
}

static long _handle = -1;

int findfirst(const char *fn, struct ffblk *fb, unsigned attrib){
	struct _finddata_t fd;
	fb->attrib = attrib;
	if(_handle >= 0){
		_findclose(_handle);
		_handle = -1;
	}
	if((_handle = _findfirst(fn, &fd)) >= 0){
		do{
			if(fd.attrib & attrib){
				fb->ff_fsize = fd.size;
				strncpy(fb->ff_name, fd.name, sizeof(fb->ff_name));
				fb->ff_name[sizeof(fb->ff_name) - 1] = 0;
				return 0;
			}
		} while(_findnext(_handle, &fd) >= 0);
	}
	_findclose(_handle);
	_handle = -1;
	return -1;
}

int findnext(struct ffblk *fb){
	struct _finddata_t fd;
	while(_findnext(_handle, &fd) >= 0){
		if(fd.attrib & fb->attrib){
			fb->ff_fsize = fd.size;
			strncpy(fb->ff_name, fd.name, sizeof(fb->ff_name));
			fb->ff_name[sizeof(fb->ff_name) - 1] = 0;
			return 0;
		}
	}
	_findclose(_handle);
	_handle = -1;
	return -1;
}

unsigned getdfree(unsigned drv_no, struct dfree *df){
	struct _diskfree_t df2;
	unsigned ret;
	ret = _getdiskfree(drv_no, &df2);
	df->df_avail = df2.avail_clusters;
	df->df_sclus = df2.sectors_per_cluster;
	df->df_bsec = df2.bytes_per_sector;
	return ret;
}

int setdisk(int drv_no){
	return _chdrive(drv_no + 1);
}

void getdate(struct date *da){
	time_t t;
	struct tm *tm;
	time(&t);
	tm = localtime(&t);
	da->da_year = tm->tm_year + 1900;
	da->da_mon = tm->tm_mon + 1;
	da->da_day = tm->tm_mday;
}

void gettime(struct time *ti){
	time_t t;
	struct tm *tm;
	time(&t);
	tm = localtime(&t);
	ti->ti_hour = tm->tm_hour;
	ti->ti_min = tm->tm_min;
	ti->ti_sec = tm->tm_sec;
	ti->ti_hund = tm->tm_sec * 100;
}

